import functools
import time
import random
from typing import Optional, TypeVar
from collections.abc import Generator
from multiprocessing import Process
import sys
import subprocess

import pytest
import numpy as np

from libertem.api import Context
from libertem.common.executor import (
    TaskCommHandler, TaskProtocol, WorkerQueue, JobCancelledError,
)
from libertem.udf.sum import SumUDF
from libertem.executor.pipelined import (
    PipelinedExecutor, WorkerPool, _order_results, pipelined_worker
)
import libertem.executor.pipelined
from libertem.udf import UDF
from libertem.common.exceptions import UDFRunCancelled
from libertem.io.dataset.memory import MemoryDataSet


class CustomException(Exception):
    pass


@pytest.fixture(scope="module")
def pipelined_ex():
    executor = None
    try:
        executor = PipelinedExecutor(
            spec=PipelinedExecutor.make_spec(cpus=range(2), cudas=[]),
            # to prevent issues in already-pinned situations (i.e. containerized
            # environments), don't pin our worker processes in testing:
            pin_workers=False,
            cleanup_timeout=5.,
        )
        yield executor
    finally:
        if executor is not None:
            print(f"pipelined_ex fixture: closing executor {id(executor)}")
            executor.close()


def test_pipelined_executor(pipelined_ex):
    executor = pipelined_ex
    ctx = Context(executor=executor)
    udf = SumUDF()
    data = np.random.randn(4, 4, 128, 128)
    ds = ctx.load("memory", data=data)
    res = ctx.run_udf(dataset=ds, udf=udf)
    assert np.allclose(
        data.sum(axis=(0, 1)),
        res['intensity'].data,
    )


def test_run_function(pipelined_ex):
    assert pipelined_ex.run_function(lambda: 42) == 42


class RaisesUDF(UDF):
    def __init__(self, exc_cls=CustomException):
        super().__init__(exc_cls=exc_cls)

    def get_result_buffers(self):
        return {
            "stuff": self.buffer(kind='nav'),
        }

    def process_frame(self, frame):
        raise self.params.exc_cls("what")


def test_udf_exception_queued(pipelined_ex):
    executor = pipelined_ex
    ctx = Context(executor=executor)

    data = np.random.randn(16, 16, 128, 128)
    ds = ctx.load("memory", data=data, num_partitions=16)

    error_udf = RaisesUDF()  # raises an error
    with pytest.raises(CustomException):  # raised by executor as expected
        ctx.run_udf(dataset=ds, udf=error_udf)

    normal_udf = SumUDF()
    ctx.run_udf(dataset=ds, udf=normal_udf)


@pytest.mark.slow
def test_default_spec():
    # make sure `.make_local` works:
    executor = None
    try:
        executor = PipelinedExecutor.make_local()

        # to at least see that something works:
        assert executor.run_function(lambda: 42) == 42
    finally:
        if executor is not None:
            executor.close()


def test_make_with():
    with Context.make_with("pipelined", cpus=1) as ctx:
        assert ctx.executor.run_function(lambda: 42) == 42


_STOP = object()

T = TypeVar('T')


def echo() -> Generator[Optional[T], T, None]:
    last = None
    while last is not _STOP:
        print(f"yielding {last}")
        last = yield last
        print(f"got {last}")
        if last is None:
            import traceback
            traceback.print_stack()


def test_order_results_in_order():
    r1 = object()
    t1 = object()
    t1id = 0

    r2 = object()
    t2 = object()
    t2id = 1

    r3 = object()
    t3 = object()
    t3id = 2

    # results come in as triples (result, task, task_id)
    def _trace_1():
        yield (r1, t1, t1id)
        yield (r2, t2, t2id)
        yield (r3, t3, t3id)

    # _order_results discards the task_id at the end
    ordered = _order_results(_trace_1())
    assert next(ordered) == (r1, t1)
    assert next(ordered) == (r2, t2)
    assert next(ordered) == (r3, t3)


def test_order_results_missing_task():
    r1 = object()
    t1 = object()
    t1id = 0

    r3 = object()
    t3 = object()
    t3id = 2

    # results come in as triples (result, task, task_id)
    def _trace_1():
        yield (r1, t1, t1id)
        yield (r3, t3, t3id)

    # _order_results discards the task_id at the end
    ordered = _order_results(_trace_1())
    assert next(ordered) == (r1, t1)
    with pytest.raises(RuntimeError):
        next(ordered)


def test_order_results_postponed_task():
    r1 = object()
    t1 = object()
    t1id = 0

    r2 = object()
    t2 = object()
    t2id = 1

    r3 = object()
    t3 = object()
    t3id = 2

    # results come in as triples (result, task, task_id)
    def _trace_1():
        yield (r1, t1, t1id)
        yield (r3, t3, t3id)
        yield (r2, t2, t2id)

    # _order_results discards the task_id at the end
    ordered = _order_results(_trace_1())
    assert next(ordered) == (r1, t1)
    assert next(ordered) == (r2, t2)
    assert next(ordered) == (r3, t3)


def test_run_function_failure(pipelined_ex):
    def _f():
        raise CustomException("this fails to run")

    with pytest.raises(CustomException) as ex_info:
        pipelined_ex.run_function(_f)

    assert ex_info.match("^this fails to run$")


def test_run_function_error():
    executor = None
    try:
        executor = PipelinedExecutor(
            spec=PipelinedExecutor.make_spec(cpus=range(2), cudas=[]),
            pin_workers=False,
        )

        def _break(a, b, c):
            raise CustomException("stuff is broken, can't do it.")

        def _do_patch_worker():
            # monkeypatch the pipelined module; this should only be run on
            # worker processes
            # XXX this completely destroys the workers ability to properly
            # function, but that's okay because it's a throwaway process pool
            # anyways:
            libertem.executor.pipelined.worker_run_function = _break

        executor.run_each_worker(_do_patch_worker)
        with pytest.raises(CustomException) as e:
            executor.run_function(lambda: 42)
        assert e.match("stuff is broken, can't do it.")
    finally:
        if executor is not None:
            executor.close()


def _broken_pipelined_worker(queues, pin, spec, span_context, early_setup):
    raise CustomException("stuff is broken, can't do it.")


def test_early_startup_error():
    """
    Simulate very early startup error, not even getting to the try/except
    that gives us error feedback via the queue.
    """
    executor = None

    # manual patching, we mock.patch doesn't work in multiprocessing
    # environments:
    original_pipelined_worker = libertem.executor.pipelined.pipelined_worker
    try:
        libertem.executor.pipelined.pipelined_worker = _broken_pipelined_worker
        with pytest.raises(RuntimeError) as e:
            executor = PipelinedExecutor(
                spec=PipelinedExecutor.make_spec(cpus=range(2), cudas=[]),
                pin_workers=False,
            )
        assert e.match("One or more workers failed to start")
    finally:
        libertem.executor.pipelined.pipelined_worker = original_pipelined_worker
        if executor is not None:
            executor.close()


def _patch_setup_device():
    def _broken_setup_device(spec, pin):
        """
        Broken version of pipelined._setup_device for error injection
        """
        raise RuntimeError("stuff is broken, can't do it.")
    libertem.executor.pipelined._setup_device = _broken_setup_device


def _slow():
    time.sleep(100)


def test_startup_error():
    """
    Simulate an error when starting up the worker, in this case we raise in
    _setup_device
    """
    executor = None
    try:
        with pytest.raises(RuntimeError) as e:
            executor = PipelinedExecutor(
                spec=PipelinedExecutor.make_spec(cpus=range(2), cudas=[]),
                pin_workers=False,
                early_setup=_patch_setup_device,
            )
        assert e.match("stuff is broken, can't do it.")
    finally:
        if executor is not None:
            executor.close()


def test_startup_timeout():
    executor = None
    try:
        with pytest.raises(RuntimeError) as e:
            executor = PipelinedExecutor(
                spec=PipelinedExecutor.make_spec(cpus=range(2), cudas=[]),
                pin_workers=False,
                early_setup=_slow,
                startup_timeout=0,
            )
        assert e.match("Timeout while starting workers")
    finally:
        if executor is not None:
            executor.close()


class FailEventuallyUDF(UDF):
    def get_result_buffers(self):
        return {
            "stuff": self.buffer(kind="nav"),
        }

    def process_partition(self, partition):
        if random.random() > 0.50:
            time.sleep(0.1)
        raise CustomException("stuff happens")


def test_failure_with_delay(pipelined_ex):
    ctx = Context(executor=pipelined_ex)
    udf = FailEventuallyUDF()
    data = np.random.randn(1, 32, 16, 16)
    ds = ctx.load("memory", data=data, num_partitions=32)
    with pytest.raises(CustomException) as e:
        ctx.run_udf(dataset=ds, udf=udf)
    assert e.match("stuff happens")


class SucceedEventuallyUDF(UDF):
    def get_result_buffers(self):
        return {
            "intensity": self.buffer(kind="nav"),
        }

    def process_partition(self, partition):
        if random.random() > 0.50:
            time.sleep(0.1)
        self.results.intensity[:] = np.sum(partition, axis=(-1, -2))


def test_success_with_delay(pipelined_ex):
    ctx = Context(executor=pipelined_ex)
    udf = SucceedEventuallyUDF()
    data = np.random.randn(1, 32, 16, 16)
    ds = ctx.load("memory", data=data, num_partitions=32)
    res = ctx.run_udf(dataset=ds, udf=udf)
    assert np.allclose(res['intensity'].data, np.sum(data, axis=(2, 3)))


def test_make_spec_multi_cuda():
    spec = PipelinedExecutor.make_spec(cpus=[0], cudas=[0, 1, 2, 2])
    assert spec == [
        {
            "device_id": 0,
            "name": "cuda-0-0",
            "device_kind": "CUDA",
            "worker_idx": 0,
            "has_cupy": False,
        },
        {
            "device_id": 1,
            "name": "cuda-1-0",
            "device_kind": "CUDA",
            "worker_idx": 1,
            "has_cupy": False,
        },
        {
            "device_id": 2,
            "name": "cuda-2-0",
            "device_kind": "CUDA",
            "worker_idx": 2,
            "has_cupy": False,
        },
        {
            "device_id": 2,
            "name": "cuda-2-1",
            "device_kind": "CUDA",
            "worker_idx": 3,
            "has_cupy": False,
        },
        {
            "device_id": 0,
            "name": "cpu-0",
            "device_kind": "CPU",
            "worker_idx": 4,
            "has_cupy": False,
        },
    ]


def test_make_spec_cpu_int():
    int_spec = PipelinedExecutor.make_spec(cpus=4, cudas=tuple(), has_cupy=True)
    range_spec = PipelinedExecutor.make_spec(cpus=range(4), cudas=tuple(), has_cupy=True)
    assert range_spec == int_spec


def test_make_spec_cuda_int():
    spec_n = 2
    cuda_spec = PipelinedExecutor.make_spec(cpus=[0], cudas=spec_n)
    num_cudas = 0
    for spec in cuda_spec:
        if spec['device_kind'] == 'CUDA':
            num_cudas += 1
    assert num_cudas == spec_n


def test_close_with_scatter():
    executor = None
    try:
        spec = PipelinedExecutor.make_spec(cpus=range(2), cudas=[])
        executor = PipelinedExecutor(spec=spec, pin_workers=False)
        with executor.scatter("huh"):
            executor.close()
    finally:
        if executor is not None:
            executor.close()


def test_exception_in_main_thread():
    executor = None
    try:
        spec = PipelinedExecutor.make_spec(cpus=range(2), cudas=[])
        executor = PipelinedExecutor(spec=spec, pin_workers=False)
        ctx = Context(executor=executor)
        udf = SucceedEventuallyUDF()
        data = np.random.randn(1, 32, 16, 16)
        ds = ctx.load("memory", data=data, num_partitions=32)
        with pytest.raises(RuntimeError):
            res_iter = ctx.run_udf_iter(dataset=ds, udf=udf)
            next(res_iter)
            res_iter.throw(RuntimeError("stuff"))
            next(res_iter)

        # here, we get a KeyError if the worker queues aren't drained:
        print("second UDF run starting")
        ctx.run_udf(dataset=ds, udf=udf)
        print("second UDF run done")
    finally:
        if executor is not None:
            executor.close()


def test_kill_pool():
    spec = PipelinedExecutor.make_spec(cpus=range(2), cudas=[])
    pool = WorkerPool(
        worker_fn=functools.partial(
            pipelined_worker,
            pin=False,
        ),
        spec=spec,
    )
    # monkey-patch: disable terminate; force use of kill
    for worker_info in pool._workers:
        worker_info.process.terminate = lambda: None
    pool.kill(timeout=0.1)


def test_term_pool():
    spec = PipelinedExecutor.make_spec(cpus=range(2), cudas=[])
    pool = WorkerPool(
        worker_fn=functools.partial(
            pipelined_worker,
            pin=False,
        ),
        spec=spec,
    )
    # hopefully this will hit the `terminate` branch:
    pool.kill(timeout=5)


def _teardown_manual():
    subprocess.run([
        sys.executable,
        '-c',
        'import time;'
        'from libertem.api import Context;'
        'ctx=Context.make_with("pipelined", cpus=1);'
        'time.sleep(0.2);'
        'ctx.close()'
    ], timeout=30)


def test_auto_teardown_1():
    subprocess.run([
        sys.executable,
        '-c',
        'import time;'
        'from libertem.api import Context;'
        'ctx=Context.make_with("pipelined", cpus=1);'
    ], timeout=30)


def test_auto_teardown_2():
    subprocess.run([
        sys.executable,
        '-c',
        'import time;'
        'import gc;'
        'from libertem.api import Context;'
        'Context.make_with("pipelined", cpus=1);'
        'gc.collect();'
    ], timeout=30)


def test_manual_teardown():
    p = Process(target=_teardown_manual, name="_teardown_manual")
    p.start()
    p.join(30)
    exitcode = p.exitcode
    if exitcode != 0:
        p.terminate()
        raise RuntimeError(f"exitcode is {exitcode}, should be 0")


class CancelledTaskCommHandler(TaskCommHandler):
    def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
        raise JobCancelledError()

    def start(self):
        pass

    def done(self):
        pass


class CancelledMemoryDataSet(MemoryDataSet):
    def get_task_comm_handler(self) -> TaskCommHandler:
        return CancelledTaskCommHandler()


def test_cancellation(pipelined_ex, default_raw):
    executor = pipelined_ex
    ctx = Context(executor=executor)

    cancel_ds = CancelledMemoryDataSet(data=np.zeros((16, 16, 16, 16)))

    with pytest.raises(UDFRunCancelled) as ex:
        ctx.run_udf(dataset=cancel_ds, udf=SumUDF())

    assert ex.match(r"^UDF run cancelled after \d+ partitions$")

    # after cancellation, the executor is still usable:
    _ = ctx.run_udf(dataset=default_raw, udf=SumUDF())



import pytest
import numpy as np

from libertem.io.dataset.memory import (
    MemPartition, MemoryDataSet, FileSet, MemoryFile
)
from libertem.common.buffers import (
    BufferWrapper, AuxBufferWrapper, reshaped_view, PlaceholderBufferWrapper,
)
from libertem.common import Shape, Slice
from libertem.udf.base import UDF

from utils import _mk_random


def test_new_for_partition():
    auxdata = _mk_random(size=(16, 16), dtype="float32")
    buf = AuxBufferWrapper(kind="nav", dtype="float32")
    buf.set_buffer(auxdata)

    dataset = MemoryDataSet(data=_mk_random(size=(16, 16, 16, 16), dtype="float32"),
                            tileshape=(7, 16, 16),
                            num_partitions=2, sig_dims=2)

    assert auxdata.shape == tuple(dataset.shape.nav)

    roi = _mk_random(size=dataset.shape.nav, dtype="bool")

    for idx, partition in enumerate(dataset.get_partitions()):
        print("partition number", idx)
        new_buf = buf.new_for_partition(partition, roi=roi)
        ps = partition.slice.get(nav_only=True)
        roi_part = roi.reshape(-1)[ps]

        assert np.product(new_buf._data.shape) == roi_part.sum()

        # old buffer stays the same:
        assert np.allclose(buf._data, auxdata.reshape(-1))
        assert buf._data_coords_global
        assert not new_buf._data_coords_global

        # new buffer is sliced to partition and has ROI applied:
        assert new_buf._data.shape[0] <= buf._data.shape[0]
        assert new_buf._data.shape[0] <= partition.shape[0]

        # let's try and manually apply the ROI to `auxdata`:
        assert np.allclose(
            new_buf._data,
            auxdata.reshape(-1)[ps][roi_part]
        )


def test_buffer_extra_shape_1():
    buffer = BufferWrapper(kind='nav', extra_shape=(2, 3))
    assert buffer._extra_shape == (2, 3)


def test_buffer_extra_shape_2():
    shape_obj = Shape(shape=(12, 13, 14, 15), sig_dims=2)
    buffer = BufferWrapper(kind='nav', extra_shape=shape_obj)
    assert buffer._extra_shape == (12, 13, 14, 15)


def test_reshaped_view():
    data = np.zeros((2, 5))
    view = data[:, :3]
    with pytest.raises(AttributeError):
        reshaped_view(view, (-1, ))
    view_2 = reshaped_view(data, (-1, ))
    view_2[0] = 1
    assert data[0, 0] == 1
    assert np.all(data[0, 1:] == 0)
    assert np.all(data[1:] == 0)


def test_result_buffer_decl():
    buf = PlaceholderBufferWrapper(kind='sig', dtype=np.float32)
    with pytest.raises(ValueError):
        # no array associated with this bufferwrapper:
        np.array(buf)


class BadMemPartition(MemPartition):
    def get_tiles(self, *args, **kwargs):
        for t in super().get_tiles(*args, **kwargs):
            sl = t.tile_slice
            origin = np.array(sl.origin)
            origin += 1
            new_sl = Slice(origin=tuple(origin), shape=sl.shape)
            t.tile_slice = new_sl
            yield t


class BadMemoryDS(MemoryDataSet):
    def get_partitions(self):
        fileset = FileSet([
            MemoryFile(
                path=None,
                start_idx=0,
                end_idx=self._image_count,
                native_dtype=self.data.dtype,
                sig_shape=self.shape.sig,
                data=self.data.reshape(self.shape.flatten_nav()),
                check_cast=self._check_cast,
            )
        ])

        for part_slice, start, stop in self.get_slices():
            yield BadMemPartition(
                meta=self._meta,
                partition_slice=part_slice,
                fileset=fileset,
                start_frame=start,
                num_frames=stop - start,
                tiledelay=self._tiledelay,
                tileshape=self.tileshape,
                force_need_decode=self._force_need_decode,
                io_backend=self.get_io_backend(),
                decoder=self.get_decoder(),
            )


class BaseSigUDF(UDF):
    def get_result_buffers(self):
        return {
            'data': self.buffer(kind='sig')
        }

    def merge(self, dest, src):
        dest.data[:] += src.data


class FrameSigUDF(BaseSigUDF):
    def process_frame(self, frame):
        self.results.data[:] += frame


class TileSigUDF(BaseSigUDF):
    def process_tile(self, tile):
        self.results.data[:] += tile.sum(axis=0)


class PartitionSigUDF(BaseSigUDF):
    def process_partition(self, partition):
        self.results.data[:] += partition.sum(axis=0)


class BaseNavUDF(UDF):
    def get_result_buffers(self):
        return {
            'data': self.buffer(kind='nav')
        }


class FrameNavUDF(BaseNavUDF):
    def process_frame(self, frame):
        self.results.data[:] = frame.sum()


class TileNavUDF(BaseNavUDF):
    def process_tile(self, tile):
        self.results.data[:] = tile.sum(axis=tuple(range(1, len(tile.shape))))


class PartitionNavUDF(BaseNavUDF):
    def process_partition(self, partition):
        self.results.data[:] = partition.sum(
            axis=tuple(range(1, len(partition.shape)))
        )


@pytest.mark.parametrize(
    'tileshape,udf_clss', (
        [
            None, (
                FrameSigUDF, TileSigUDF, PartitionSigUDF, FrameNavUDF,
                TileNavUDF, PartitionNavUDF
            )
        ],
        [(7, 16, 16), (FrameSigUDF, TileSigUDF, FrameNavUDF, TileNavUDF)],
        [(7, 7, 7), (TileSigUDF, TileNavUDF)],
        [(16, 8, 16), (TileSigUDF, TileNavUDF)],
        [
            (8*16, 16, 16), (
                FrameSigUDF, TileSigUDF, PartitionSigUDF, FrameNavUDF,
                TileNavUDF, PartitionNavUDF
            )
        ],
    )
)
def test_buffer_slices(lt_ctx, tileshape, udf_clss):
    data = _mk_random(size=(16, 16, 16, 16))
    bad_ds = BadMemoryDS(
        data=data,
        tileshape=tileshape,
        num_partitions=2,
        sig_dims=2
    )
    ds = MemoryDataSet(
        data=data,
        tileshape=tileshape,
        num_partitions=2,
        sig_dims=2
    )

    for udf_cls in udf_clss:
        _ = lt_ctx.run_udf(dataset=ds, udf=udf_cls())
        with pytest.raises(Exception) as exc_info:
            _ = lt_ctx.run_udf(dataset=bad_ds, udf=udf_cls())
        assert exc_info.errisinstance((AssertionError, IndexError))



import os
import pytest
import numpy as np

from libertem.io.dataset.raw import RawFileDataSet, RawFileSet, RawFile


RUNNER_OS = os.environ.get("RUNNER_OS", "").lower().strip()
RUNNER_ARCH = os.environ.get("RUNNER_ARCH", "").lower().strip()

try:
    import resource
    _, hard_lim = resource.getrlimit(resource.RLIMIT_NOFILE)
    pytestmark = pytest.mark.skipif(hard_lim < 2 ** 15, reason="hard file limit is too low")  # NOQA
except ModuleNotFoundError:
    # Not available on Windows
    pass


@pytest.fixture(scope='session')
def generate_many_files(tmpdir_factory):
    # Generates 16k 8x8 uint8 frames
    # Combined with the below 2-partition dataset and
    # the two-cpu local_cluster_ctx this is likely to trigger file
    # limits on a system where this is not set to the hard limit globally
    datadir = tmpdir_factory.mktemp('many_raw_files')
    frame = np.arange(64).astype(np.uint8)
    nav_shape = (128, 100)
    for idx in range(np.prod(nav_shape)):
        frame.tofile(os.path.join(datadir, f'f{idx:>05d}.raw'))
    yield datadir


class ManyRawFileDataSetMock(RawFileDataSet):
    """
    A basic multi-file raw dataset that reads every .raw file
    in a folder, assuming one image per frame and all files
    are of an identical size and unordered
    """
    def _get_raw_files(self):
        with os.scandir(self._path) as it:
            for entry in it:
                if entry.is_file() and entry.name.endswith('.raw'):
                    yield entry

    def _get_filesize(self):
        return sum(p.stat().st_size for p in self._get_raw_files())

    def _get_fileset(self):
        return RawFileSet([
            RawFile(
                path=p.path,
                start_idx=idx,
                end_idx=idx + 1,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
            )
            for idx, p in enumerate(self._get_raw_files())])

    def get_num_partitions(self):
        return 2


@pytest.mark.skipif(
    RUNNER_OS.startswith('macos') and RUNNER_ARCH.startswith("arm"),
    reason="Hard limit not compatible on macos-14 runner"
)
def test_many_files_read(local_cluster_ctx, generate_many_files):
    """
    Tests if we can load 8k files per process in a Dask cluster
    on this system. Depending on the system configuration
    this test might be a no-op.
    """
    ds = ManyRawFileDataSetMock(
        path=generate_many_files,
        nav_shape=(128, 100),
        sig_shape=(8, 8),
        dtype=np.uint8,
    )
    ds.initialize(local_cluster_ctx.executor)

    sum_a = local_cluster_ctx.create_sum_analysis(ds)
    local_cluster_ctx.run(sum_a)



import pytest
import numpy as np

from libertem.io.dataset.memory import MemoryDataSet

from utils import _naive_mask_apply, _mk_random


@pytest.fixture
def ds_random():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2,
    )
    return dataset


def test_disk_1(lt_ctx, ds_random):
    analysis = lt_ctx.create_disk_analysis(dataset=ds_random, cx=8, cy=8, r=5)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_random.data)
    # TODO: test the actual mask shape, too
    # at least mask application should match:
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_disk_defaults(lt_ctx, ds_random):
    analysis = lt_ctx.create_disk_analysis(dataset=ds_random)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_random.data)
    # TODO: test the actual mask shape, too
    # at least mask application should match:
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_ring_1(lt_ctx, ds_random):
    analysis = lt_ctx.create_ring_analysis(dataset=ds_random, cx=8, cy=8, ri=5, ro=8)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_random.data)
    assert results.intensity.raw_data.shape == (16, 16)
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_ring_3d_ds(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2,
    )
    analysis = lt_ctx.create_ring_analysis(dataset=dataset, cx=8, cy=8, ri=5, ro=8)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], dataset.data.reshape((16, 16, 16, 16)))
    assert results.intensity.raw_data.shape == (16 * 16,)
    assert np.allclose(
        results.intensity.raw_data.reshape((16, 16)),
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data.reshape((16, 16)),
        expected,
    )


def test_ring_defaults(lt_ctx, ds_random):
    analysis = lt_ctx.create_ring_analysis(dataset=ds_random)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_random.data)
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_point_1(lt_ctx, ds_random):
    analysis = lt_ctx.create_point_analysis(dataset=ds_random, x=8, y=8)
    results = lt_ctx.run(analysis)
    mask = np.zeros((16, 16))
    mask[8, 8] = 1
    expected = _naive_mask_apply([mask], ds_random.data)
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_point_3d_ds(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2,
    )
    analysis = lt_ctx.create_point_analysis(dataset=dataset, x=8, y=8)
    results = lt_ctx.run(analysis)
    mask = np.zeros((16, 16))
    mask[8, 8] = 1
    expected = _naive_mask_apply([mask], dataset.data.reshape((16, 16, 16, 16)))
    assert results.intensity.raw_data.shape == (16 * 16,)
    assert np.allclose(
        results.intensity.raw_data.reshape((16, 16)),
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data.reshape((16, 16)),
        expected,
    )


def test_point_defaults(lt_ctx, ds_random):
    analysis = lt_ctx.create_point_analysis(dataset=ds_random)
    results = lt_ctx.run(analysis)
    mask = np.zeros((16, 16))
    mask[8, 8] = 1
    expected = _naive_mask_apply([mask], ds_random.data)
    assert results.intensity.raw_data.shape == (16, 16)
    assert np.allclose(
        results.intensity.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        expected,
    )


def test_disk_complex(lt_ctx, ds_complex):
    analysis = lt_ctx.create_disk_analysis(dataset=ds_complex)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_complex.data)
    assert np.allclose(
        results.intensity_complex.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        np.abs(results.intensity_complex),
    )
    assert np.allclose(
        results.intensity.raw_data,
        np.abs(results.intensity_complex),
    )


def test_ring_complex(lt_ctx, ds_complex):
    analysis = lt_ctx.create_ring_analysis(dataset=ds_complex)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_complex.data)
    assert np.allclose(
        results.intensity_complex.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        np.abs(results.intensity_complex),
    )
    assert np.allclose(
        results.intensity.raw_data,
        np.abs(results.intensity_complex),
    )


def test_point_complex(lt_ctx, ds_complex):
    analysis = lt_ctx.create_point_analysis(dataset=ds_complex)
    results = lt_ctx.run(analysis)
    mask = analysis.get_mask_factories()[0]()
    expected = _naive_mask_apply([mask], ds_complex.data)
    assert np.allclose(
        results.intensity_complex.raw_data,
        expected,
    )
    assert np.allclose(
        results.intensity_log.raw_data,
        np.abs(results.intensity_complex),
    )
    assert np.allclose(
        results.intensity.raw_data,
        np.abs(results.intensity_complex),
    )



import json

import pytest
import websockets

from utils import assert_msg
from aio_utils import create_connection

from libertem.executor.inline import InlineJobExecutor
from libertem.executor.base import AsyncAdapter

pytestmark = [pytest.mark.web_api]


def _get_raw_params(path):
    return {
        "dataset": {
            "params": {
                "type": "RAW",
                "path": path,
                "dtype": "float32",
                "sig_shape": [128, 128],
                "enable_direct": False,
                "nav_shape": [16, 16],
                "sync_offset": 0
            }
        }
    }


@pytest.mark.asyncio
async def test_load_raw_success(
    default_raw, base_url, http_client, local_cluster_url, default_token
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    raw_path = default_raw._path

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params(raw_path)
    async with http_client.put(ds_url, json=ds_data) as resp:
        assert resp.status == 200
        resp_json = await resp.json()
        assert_msg(resp_json, 'CREATE_DATASET')
        for k in ds_data['dataset']['params']:
            assert ds_data['dataset']['params'][k] == resp_json['details']['params'][k]


@pytest.mark.asyncio
async def test_load_raw_with_backends(
    default_raw, base_url, http_client, local_cluster_url, default_token
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    raw_path = default_raw._path

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params(raw_path)
    for backend in ["mmap", "buffered"]:
        ds_data["dataset"]["params"]["io_backend"] = backend
        async with http_client.put(ds_url, json=ds_data) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'CREATE_DATASET')
            for k in ds_data['dataset']['params']:
                assert ds_data['dataset']['params'][k] == resp_json['details']['params'][k]
            assert ds_data["dataset"]["params"]["io_backend"] == backend
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_load_raw_fail(base_url, http_client, local_cluster_url, default_token):
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params("/does/not/exist/")
    async with http_client.put(ds_url, json=ds_data) as resp:
        assert resp.status == 200
        resp_json = await resp.json()
        assert_msg(resp_json, 'CREATE_DATASET_ERROR', status='error')
        assert resp_json['dataset'] == uuid
        assert (
            "No such file or directory" in resp_json['msg']
            or "The system cannot find the path specified" in resp_json['msg']
        )


@pytest.mark.asyncio
async def test_dataset_delete(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    raw_path = default_raw._path

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params(raw_path)

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        async with http_client.put(ds_url, json=ds_data) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'CREATE_DATASET')

        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_initial_state_after_reconnect(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    raw_path = default_raw._path

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params(raw_path)

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        async with http_client.put(ds_url, json=ds_data) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'CREATE_DATASET')

    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')
        assert initial_msg["jobs"] == []
        assert len(initial_msg["datasets"]) == 1
        assert initial_msg["datasets"][0]["id"] == uuid
        assert initial_msg["datasets"][0]["params"] == {
            'sig_shape': [128, 128],
            "enable_direct": False,
            'dtype': 'float32',
            'path': raw_path,
            'nav_shape': [16, 16],
            'shape': [16, 16, 128, 128],
            'type': 'RAW',
            'sync_offset': 0
        }
        assert len(initial_msg["datasets"][0]["diagnostics"]) == 7


@pytest.mark.asyncio
async def test_prime_cache(
    shared_state, default_raw, base_url, http_client, server_port, local_cluster_url,
    default_token,
):
    # first, connect to get the state
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    executor = InlineJobExecutor()

    pool = AsyncAdapter.make_pool()
    executor = AsyncAdapter(wrapped=executor, pool=pool)
    conn_details = {
        'connection': {
            'type': 'local',
            'numWorkers': 1,
            'cudas': [],
        }
    }
    await shared_state.executor_state.set_executor(executor, conn_details)

    raw_path = default_raw._path

    uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, uuid, default_token,
    )
    ds_data = _get_raw_params(raw_path)

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        async with http_client.put(ds_url, json=ds_data) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'CREATE_DATASET')

    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')
        assert initial_msg["jobs"] == []
        assert len(initial_msg["datasets"]) == 1
        assert initial_msg["datasets"][0]["id"] == uuid
        assert initial_msg["datasets"][0]["params"] == {
            'sig_shape': [128, 128],
            "enable_direct": False,
            'dtype': 'float32',
            'path': raw_path,
            'nav_shape': [16, 16],
            'shape': [16, 16, 128, 128],
            'type': 'RAW',
            'sync_offset': 0
        }
        assert len(initial_msg["datasets"][0]["diagnostics"]) == 7



from io import SEEK_SET
import math
import os
import re
from typing import IO, TYPE_CHECKING, NamedTuple, Optional
import numpy as np
from glob import glob, escape

from libertem.common.math import prod, make_2D_square
from libertem.common import Shape
from libertem.common.executor import JobExecutor
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, DataSetException, DataSetMeta,
    BasePartition, File, FileSet, IOBackend,
)

if TYPE_CHECKING:
    from numpy import typing as nt


class TVIPSDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/TVIPSDatasetParams.schema.json",
        "title": "TVIPSDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "TVIPS"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
        },
        "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        for k in ["nav_shape", "sig_shape", "sync_offset"]:
            if k in raw_data:
                data[k] = raw_data[k]
        return data


SERIES_HEADER_SIZE = 256

series_header_dtype = [
    ('ISize', 'i4'),            # The size of the series header (always 256)
    ('IVersion', 'i4'),         # The version of the file (1 or 2)
    ('IXDim', 'i4'),            # The x dimension of all images (width)
    ('IYDim', 'i4'),            # The y dimension of all images (height)
    ('IBPP', 'i4'),             # The number of bits per pixel (8 or 16)
    ('IXOff', 'i4'),            # The camera X offset of the image
    ('IYOff', 'i4'),            # The camera Y offset of the image
    ('IXBin', 'i4'),            # The camera X binning
    ('IYBin', 'i4'),            # The camera Y binning
    ('IPixelSize', 'i4'),       # The pixelsize in nm
    ('IHT', 'i4'),              # The high tension in kV
    ('IMagTotal', 'i4'),        # The total magnification including MagPost and MagCor factors
    ('IImgHeaderBytes', 'i4'),  # The size in bytes of the image headers (version 2 only)
    # 204 unused bytes follow
]


image_header_v2_dtype = [
    ('ICounter', 'u4'),                 # image counter, continues through all files
    ('ITime', 'u4'),                    # unix time stamp
    ('IMS', 'u4'),                      # timestamp milliseconds
    ('LUT_Index', 'u4'),                # LUT index (?)
    ('Faraday', 'float32'),             # faraday cup value (unit?)
    ('TEM_Mag', 'u4'),                  # magnification (unit?)
    ('TEM_Mag_mode', 'u4'),             # magnification mode (1=imaging, 2=diffraction)
    ('TEM_Stage_x', 'float32'),         # stage X in nm
    ('TEM_Stage_y', 'float32'),         # stage Y in nm
    ('TEM_Stage_z', 'float32'),         # stage Z in nm
    ('TEM_Stage_alpha', 'float32'),     # in degree
    ('TEM_Stage_beta', 'float32'),      # in degree
    ('Index_of_rotator', 'u4'),         # ?
    ('DENS_T_measure', 'float32'),
    ('DENS_T_setpoint', 'float32'),
    ('DENS_Power', 'float32'),
    ('TEM_Obj_current', 'float32'),     # unit?
    ('Scan_x', 'float32'),
    ('Scan_y', 'float32'),
    ('DENS_Bias_U_setpoint', 'float32'),
    ('DENS_Bias_U_value', 'float32'),
    ('DENS_Bias_I_setpoint', 'float32'),
    ('DENS_Bias_I_value', 'float32'),
    ('DENS_Bias_E_setpoint', 'float32'),
    ('DENS_Bias_R', 'float32'),
    ('DENS_Bias_limit_U', 'float32'),  # compliance limit
    ('DENS_Bias_limit_I', 'float32'),  # compliance limit
]


class SeriesHeader(NamedTuple):
    version: int
    xdim: int
    ydim: int
    xbin: int
    ybin: int
    bpp: int
    pixel_size_nm: int
    high_tension_kv: int
    mag_total: int
    frame_header_bytes: int


def read_series_header(path: str) -> SeriesHeader:
    with open(path, 'rb') as f:
        arr = np.fromfile(f, dtype=series_header_dtype, count=1)
    version = int(arr['IVersion'][0])
    if version not in [1, 2]:
        raise DataSetException(f"Unknown TVIPS header version: {version}")
    size = int(arr['ISize'][0])
    if size != SERIES_HEADER_SIZE:
        raise DataSetException(
            f"Invalid header size {size}, should be 256. Maybe not a TVIPS file?"
        )
    bpp = int(arr['IBPP'][0])
    if bpp not in [8, 16]:
        raise DataSetException(
            f"unknown bpp value: {bpp} (should be either 8 or 16)"
        )
    img_header_bytes = int(arr['IImgHeaderBytes'][0])
    if version == 1:
        img_header_bytes = 12
    return SeriesHeader(
        version=int(arr['IVersion'][0]),
        xdim=int(arr['IXDim'][0]),
        ydim=int(arr['IYDim'][0]),
        xbin=int(arr['IXBin'][0]),
        ybin=int(arr['IYBin'][0]),
        bpp=bpp,
        pixel_size_nm=int(arr['IPixelSize'][0]),
        high_tension_kv=int(arr['IHT'][0]),
        mag_total=int(arr['IMagTotal'][0]),
        frame_header_bytes=img_header_bytes,
    )


def frames_in_file(path: str, series_header: SeriesHeader) -> int:
    filesize = os.stat(path).st_size
    file_header = 0
    if _get_suffix(path) == 0:
        file_header = SERIES_HEADER_SIZE
    filesize -= file_header
    total_size_per_frame = series_header.frame_header_bytes + (
        series_header.bpp // 8 * series_header.xdim * series_header.ydim
    )
    rest = filesize % total_size_per_frame
    assert rest == 0, f"found a rest of {rest}, corrupted file?"
    return filesize // total_size_per_frame


def _pattern(path: str) -> str:
    path, ext = os.path.splitext(path)
    ext = ext.lower()
    if ext == '.tvips':
        pattern = "%s*.tvips" % (
            re.sub(r'[0-9]+$', '', escape(path))
        )
    else:
        raise DataSetException("unknown extension")
    return pattern


def get_image_count_and_sig_shape(path: str) -> tuple[int, tuple[int, int]]:
    fns = get_filenames(path)
    count = 0
    series_header = read_series_header(fns[0])
    for path in fns:
        count += frames_in_file(path, series_header)
    sig_shape = (series_header.ydim, series_header.xdim)
    return count, sig_shape


MAX_SCAN_IDX = 4096  # we only check until this index for the beginning of the scan


def _image_header_for_idx(f: IO[bytes], series_header: SeriesHeader, idx: int) -> np.ndarray:
    image_size_bytes = series_header.bpp // 8 * series_header.xdim * series_header.ydim
    skip_size = series_header.frame_header_bytes + image_size_bytes
    offset = SERIES_HEADER_SIZE + idx * skip_size
    f.seek(offset, SEEK_SET)
    return np.fromfile(f, dtype=image_header_v2_dtype, count=1)  # type:ignore


def _scan_for_idx(f: IO[bytes], series_header: SeriesHeader, idx: int) -> tuple[int, int]:
    arr = _image_header_for_idx(f, series_header, idx)
    # this assumes integer scan coordinates:
    scan_y = int(arr['Scan_y'][0])
    scan_x = int(arr['Scan_x'][0])
    scan = (scan_y, scan_x)
    return scan


class DetectionError(Exception):
    pass


def detect_shape(path: str) -> tuple[int, tuple[int, ...]]:
    series_header = read_series_header(path)

    if series_header.version != 2:
        raise DetectionError(
            "unknown series header version, can only detect shape from v2"
        )

    count, _ = get_image_count_and_sig_shape(path)
    filenames = get_filenames(path)
    first_file = filenames[0]
    sync_offset = 0

    with open(first_file, "rb") as f:
        idx = 0
        last_was_zero = False
        found_offset = False
        while idx < MAX_SCAN_IDX and idx < count:
            scan = _scan_for_idx(f, series_header, idx)
            if last_was_zero and scan == (0, 1):
                sync_offset = idx - 1
                found_offset = True
                break
            if scan == (0, 0):
                last_was_zero = True
            idx += 1

        if not found_offset:
            raise DetectionError("Could not auto-detect sync_offset")

        # continue where we left off and search for max(scan_x):
        max_x = 0  # scan positions start at 0, so our shape is (y, max_x + 1)
        found_shape = False
        while idx < MAX_SCAN_IDX and idx < count:
            scan = _scan_for_idx(f, series_header, idx)
            # assume monotonously increasing values
            max_x = max(max_x, scan[1])
            if scan[1] < max_x:
                found_shape = True
                break
            idx += 1

    shape: tuple[int, ...]
    if found_shape:
        shape = (int(math.floor((count - sync_offset) / (max_x + 1))), max_x + 1)
    else:
        shape = (count,)

    return sync_offset, shape


def _get_suffix(path: str) -> int:
    path, ext = os.path.splitext(path)
    # according to the docs, the suffix is always an underscore followed
    # by a three digit number with leading zeros:
    return int(path[-3:])


def get_filenames(path: str) -> list[str]:
    return list(sorted(glob(_pattern(path)), key=_get_suffix))


class TVIPSDataSet(DataSet):
    """
    Read data from one or more .tvips files. You can specify the path to any
    file that is part of a set - the whole data set will be loaded. We will try
    to guess :code:`nav_shape` and :code:`sync_offset` from the image headers
    for 4D STEM data, but you may need to specify these parameters in case the
    guessing logic fails.

    .. versionadded:: 0.9.0

    Examples
    --------

    >>> ds = ctx.load(
    ...     "tvips",
    ...     path="./path/to/file_000.tvips",
    ...     nav_shape=(16, 16)
    ... )  # doctest: +SKIP

    Parameters
    ----------

    path: str
        Path to the file

    nav_shape: tuple of int
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int
        Common case: (height, width); but can be any dimensionality

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
        If not given, we try to automatically determine the sync_offset from
        the scan metadata in the image headers.
    """
    def __init__(
        self,
        path,
        nav_shape: Optional[tuple[int, ...]] = None,
        sig_shape: Optional[tuple[int, ...]] = None,
        sync_offset: Optional[int] = None,
        io_backend: Optional[IOBackend] = None,
    ):
        super().__init__(io_backend=io_backend)
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        self._path = path
        self._filesize = None
        self._files: Optional[list[str]] = None
        self._frame_counts: dict[str, int] = {}
        self._series_header: Optional[SeriesHeader] = None

    def initialize(self, executor: JobExecutor):
        self._filesize = executor.run_function(self._get_filesize)
        files = executor.run_function(get_filenames, self._path)

        # The series header is contained in the first file:
        self._series_header = executor.run_function(read_series_header, files[0])

        for fname in files:
            self._frame_counts[fname] = executor.run_function(
                frames_in_file,
                fname,
                self._series_header
            )

        self._files = files

        try:
            sync_offset_detected, nav_shape_detected = executor.run_function(
                detect_shape, self._path
            )
            if self._sync_offset is None:
                self._sync_offset = sync_offset_detected
        except DetectionError:
            sync_offset_detected = None
            nav_shape_detected = None
            if self._sync_offset is None:
                self._sync_offset = 0

        # The total number of frames is not contained in a header, so we need
        # to calculate it from the file sizes:
        image_count = sum(
            executor.run_function(frames_in_file, fname, self._series_header)
            for fname in files
        )

        raw_dtype: "nt.DTypeLike"
        if self._series_header.bpp == 8:
            raw_dtype = np.uint8
        elif self._series_header.bpp == 16:
            raw_dtype = np.uint16

        nav_shape: tuple[int, ...]
        if self._nav_shape is None and nav_shape_detected is not None:
            nav_shape = nav_shape_detected
        elif self._nav_shape is None and nav_shape_detected is None:
            nav_shape = (image_count,)
        elif self._nav_shape is not None:
            nav_shape = self._nav_shape
        else:
            raise RuntimeError("should not happen")  # logic and all that good stuff...

        self._image_count = image_count
        self._nav_shape_product = prod(nav_shape)
        image_size = (self._series_header.ydim, self._series_header.xdim)

        sig_shape: tuple[int, ...]
        if self._sig_shape is None:
            sig_shape = image_size
        elif prod(self._sig_shape) != prod(image_size):
            raise DataSetException(
                "sig_shape must be of size: %s" % prod(image_size)
            )
        else:
            sig_shape = self._sig_shape

        # FIXME: reshaping self._sig_shape, self._nav_shape
        shape = Shape(
            nav_shape + sig_shape,
            sig_dims=2,
        )

        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=raw_dtype,
            sync_offset=self._sync_offset,
            image_count=image_count,
        )
        return self

    def _get_filesize(self):
        files = get_filenames(self._path)
        return sum(
            os.stat(fname).st_size
            for fname in files
        )

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    @classmethod
    def get_msg_converter(cls):
        return TVIPSDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"tvips"}

    @classmethod
    def detect_params(cls, path, executor):
        pathlow = path.lower()
        if pathlow.endswith(".tvips"):
            image_count, sig_shape = executor.run_function(get_image_count_and_sig_shape, path)
            try:
                sync_offset, nav_shape = executor.run_function(detect_shape, path)
            except DetectionError:
                sync_offset = 0
                nav_shape = make_2D_square((image_count,))
        else:
            return False
        return {
            "parameters": {
                "path": path,
                "nav_shape": nav_shape,
                "sig_shape": sig_shape,
                "sync_offset": sync_offset,
            },
            "info": {
                "image_count": image_count,
                "native_sig_shape": sig_shape,
            }
        }

    def get_diagnostics(self):
        header = self._series_header
        return [
            {"name": "Bits per pixel",
             "value": str(header.bpp)},
            {"name": "High tension (kV)",
             "value": str(header.high_tension_kv)},
            {"name": "Pixel size (nm)",
             "value": str(header.pixel_size_nm)},
            {"name": "Binning (x)",
             "value": str(header.xbin)},
            {"name": "Binning (y)",
             "value": str(header.ybin)},
            {"name": "File Format Version",
             "value": str(header.version)},
        ]

    def _get_fileset(self):
        filenames = self._files
        series_header = self._series_header
        start_idx = 0
        files = []
        for fname in filenames:
            num_frames = self._frame_counts[fname]
            files.append(
                File(
                    path=fname,
                    file_header=SERIES_HEADER_SIZE if _get_suffix(fname) == 0 else 0,
                    start_idx=start_idx,
                    end_idx=start_idx + num_frames,
                    sig_shape=self.shape.sig,
                    native_dtype=self._meta.raw_dtype,
                    frame_header=series_header.frame_header_bytes,
                )
            )
            start_idx += num_frames
        return FileSet(files, frame_header_bytes=series_header.frame_header_bytes)

    def check_valid(self):
        try:
            fileset = self._get_fileset()
            backend = self.get_io_backend().get_impl()
            with backend.open_files(fileset):
                return True
        except (OSError, ValueError) as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path": self._path,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<TVIPSDataSet shape={self.shape}>"



import pytest
import numpy as np
from numpy.testing import assert_allclose
from sparseconverter import NUMPY
from libertem.common.shape import Shape
from libertem.common.slice import Slice
from libertem.common.math import flat_nonzero
from libertem.io.dataset.base import TilingScheme
from libertem.io.dataset.base.backend_mmap import MMapBackendImpl, MMapFile

from libertem.io.dataset.mib import (
    encode_r1,
    encode_r6,
    encode_r12,
    HeaderDict,
    MIBDecoder,
    MIBFile,
    MIBFileSet,
    mib_2x2_get_read_ranges,
)


def encode_quad(encode, data, bits_per_pixel, with_headers=False):
    """
    Parameters
    ==========
    encode : Callable
        One of the `encode_r*` functions

    data : np.ndarray
        The array that should be encoded, with dtype int, shape (-1, 512, 512)

    bits_per_pixel : int
        One of 1, 8, 16 - the bits per pixels padded to byte boudaries

    with_headers : bool
        Will insert some random data between the frames, not real headers.
    """
    shape = data.shape
    num_frames = shape[0]
    # typically the header size for quad data, but doesn't really matter,
    # as we don't generate a "true" header, but just random padding at
    # the beginning of each frame.
    header_bytes = 768
    assert len(shape) == 3  # decoding multiple frames at once
    enc_bytes_per_frame = shape[1] * shape[2] // 8 * bits_per_pixel
    x_shape_px = 256 // 8 * bits_per_pixel

    encoded = np.zeros(
        data.size // 8 * bits_per_pixel + shape[0] * header_bytes,
        dtype=np.uint8
    )
    encoded = encoded.reshape((-1, enc_bytes_per_frame + header_bytes))

    # encoders only do one frame per call:
    for i in range(shape[0]):
        encoded[i, :header_bytes] = np.random.randint(0, 0x100, header_bytes)

        # reshape destination buffer to allow convenient row-based assignment:
        # dest = [4 | 3 | 2 | 1]
        dest = encoded[i, header_bytes:].reshape((256, -1))
        assert dest.shape == (256, 4 * x_shape_px)

        src = data[i]
        src_half = src.shape[0] // 2, src.shape[1] // 2

        q1 = src[:src_half[0], :src_half[1]]
        encode(inp=q1, out=dest[:, 3 * x_shape_px:])

        q2 = src[:src_half[0], src_half[1]:]
        encode(inp=q2, out=dest[:, 2 * x_shape_px:3 * x_shape_px])

        # q3/q4 flipped in y direction
        q3 = src[src_half[0]:, :src_half[1]][::-1, ::-1]
        encode(inp=q3, out=dest[:, 1 * x_shape_px:2 * x_shape_px])
        q4 = src[src_half[0]:, src_half[1]:][::-1, ::-1]
        encode(inp=q4, out=dest[:, 0 * x_shape_px:1 * x_shape_px])

    if with_headers:
        return encoded
    else:
        encoded_data = encoded.reshape((num_frames, -1,))[:, header_bytes:].reshape(
            (num_frames, data.shape[1], -1)
        )
        return encoded_data


class InMemoryFile(MMapFile):
    """
    For testing purposes, this `MMapFile` gets the mmap-like objects
    from the underlying `desc`.
    """
    def __init__(self, path, desc):
        super().__init__(path, desc)
        self._handle = None
        encoded_data = desc.data
        assert encoded_data.dtype == np.uint8
        self._mmap = encoded_data.reshape((-1,))
        self._array = self._mmap

    def open(self):
        return self

    def close(self):
        pass  # do nothing

    @property
    def handle(self):
        return None


class MMapBackendImplInMem(MMapBackendImpl):
    FILE_CLS = InMemoryFile


def encode_roundtrip_quad(
    encode, bits_per_pixel, input_data=None, dataset_shape=None, tileshape=None,
    start_at_frame=2, stop_before_frame=6
):
    if dataset_shape is None:
        # make some read ranges:
        dataset_shape = (6, 512, 512)
    dataset_shape = Shape(dataset_shape, sig_dims=2)
    if tileshape is None:
        tileshape = (2, 128, 512)
    tiling_scheme = TilingScheme.make_for_shape(
        dataset_shape=dataset_shape,
        tileshape=Shape(tileshape, sig_dims=2),
    )
    sync_offset = 0
    roi = None

    frame_header_bytes = 768

    image_size_bytes = dataset_shape.sig.size * bits_per_pixel // 8

    if bits_per_pixel in (1, 8):
        native_dtype = np.uint8
    elif bits_per_pixel == 16:
        native_dtype = np.uint16

    fields: HeaderDict = {
        'header_size_bytes': frame_header_bytes,
        'dtype': native_dtype,
        'mib_dtype': 'R64',
        'mib_kind': 'r',

        # remove padding from `bits_per_pixel`
        'bits_per_pixel': {1: 1, 8: 6, 16: 12}[bits_per_pixel],
        'image_size': (512, 512),
        'image_size_bytes': image_size_bytes,
        'sequence_first_image': 1,
        'filesize': dataset_shape.nav.size * (image_size_bytes + frame_header_bytes),
        'num_images': dataset_shape.nav.size,
        'num_chips': 4,
        'sensor_layout': (2, 2),
    }

    file = MIBFile(
        path="",
        start_idx=0,
        end_idx=dataset_shape.nav.size,
        native_dtype=native_dtype,
        sig_shape=dataset_shape.sig,
        frame_header=frame_header_bytes,
        file_header=0,
        header=fields,
    )

    fileset = MIBFileSet(files=[file], header=fields, frame_header_bytes=frame_header_bytes)

    backend = MMapBackendImplInMem()

    max_value = (1 << bits_per_pixel) - 1
    if input_data is None:
        data_full = np.random.randint(0, max_value + 1, tuple(dataset_shape.flatten_nav()))
        # make sure min/max values are indeed hit:
        data_full.reshape((-1,))[0] = max_value
        data_full.reshape((-1,))[-1] = 0
        assert np.max(data_full) == max_value
        assert np.min(data_full) == 0
    else:
        data_full = input_data.reshape(dataset_shape.flatten_nav())
    data = data_full[start_at_frame:stop_before_frame]

    # we need headers in-between, in contrast to the frame-by-frame decoding, the decoder
    # expects contiguous input data and we can't slice them away beforehand:
    encoded_data = encode_quad(encode, data_full, bits_per_pixel, with_headers=True)
    decoded = np.zeros_like(data)

    # that's the "interface" we made up for the in-mem mmap file above:
    file.data = encoded_data

    # wrapping the numba decoder function:
    decoder = MIBDecoder(header=fields)

    outer_slice = Slice(
        origin=(start_at_frame, 0, 0),
        shape=dataset_shape.flatten_nav(),
    )

    read_ranges = fileset.get_read_ranges(
        start_at_frame=start_at_frame,
        stop_before_frame=stop_before_frame,
        dtype=native_dtype,
        tiling_scheme=tiling_scheme,
        sync_offset=sync_offset,
        roi=roi,
    )

    for tile in backend.get_tiles(
        tiling_scheme=tiling_scheme,
        fileset=fileset,
        read_ranges=read_ranges,
        roi=roi,
        native_dtype=np.uint8,
        read_dtype=np.float32,
        decoder=decoder,
        sync_offset=0,
        corrections=None,
        array_backend=NUMPY,
    ):
        slice_shifted = tile.tile_slice.shift(outer_slice)
        decoded[slice_shifted.get()] = tile.data.reshape(tile.tile_slice.shape)

    assert_allclose(data, decoded)
    return data, decoded


@pytest.mark.with_numba
@pytest.mark.parametrize(
    'encode,bits_per_pixel', [
        (encode_r1, 1),
        (encode_r6, 8),
        (encode_r12, 16),
    ],
)
@pytest.mark.slow
def test_encode_roundtrip_quad(encode, bits_per_pixel):
    data, decoded = encode_roundtrip_quad(
        encode, bits_per_pixel, dataset_shape=(1, 512, 512), tileshape=(1, 512, 512),
        start_at_frame=0, stop_before_frame=1,
    )
    assert_allclose(data, decoded)


@pytest.mark.slow
@pytest.mark.parametrize(
    'encode,bits_per_pixel', [
        (encode_r1, 1),
        (encode_r6, 8),
        (encode_r12, 16),
    ],
)
def test_encode_roundtrip_quad_slow(encode, bits_per_pixel):
    """
    This version is built to be a bit more thorough by taking in larger amounts
    of data and making sure it is still decoded properly. This is not marked `with_numba`
    as the non-numba code runs really slow.
    """
    data, decoded = encode_roundtrip_quad(encode, bits_per_pixel)
    assert_allclose(data, decoded)


@pytest.mark.parametrize('bits_per_pixel', (1, 8, 16))
def test_readranges_quad(bits_per_pixel):
    # make some read ranges:
    dataset_shape = Shape((4, 4, 512, 512), sig_dims=2)
    tiling_scheme = TilingScheme.make_for_shape(
        dataset_shape=dataset_shape,
        tileshape=Shape((2, 32, 512), sig_dims=2),
    )
    sync_offset = 0
    start_at_frame = 2
    stop_before_frame = 6
    roi = None
    # fileset_arr with one "file":
    fileset_arr = np.zeros((1, 4), dtype=np.int64)

    frame_header_bytes = 768
    frame_footer_bytes = 0

    # This is for formats like DM where the offset in each file can be different.
    # In this case, we only have the per-frame header:
    file_header_bytes = 0

    # (start_idx, end_idx, file_idx, file_header_bytes)
    fileset_arr[0] = (0, dataset_shape.nav.size, 0, file_header_bytes)

    roi_nonzero = None
    if roi is not None:
        roi_nonzero = flat_nonzero(roi).astype(np.int64)

    kwargs = dict(
        start_at_frame=start_at_frame,
        stop_before_frame=stop_before_frame,
        roi_nonzero=roi_nonzero,
        depth=tiling_scheme.depth,
        slices_arr=tiling_scheme.slices_array,
        fileset_arr=fileset_arr,
        sig_shape=tuple(tiling_scheme.dataset_shape.sig),
        sync_offset=sync_offset,
        bpp=bits_per_pixel,
        frame_header_bytes=frame_header_bytes,
        frame_footer_bytes=frame_footer_bytes,
    )
    read_ranges = mib_2x2_get_read_ranges(**kwargs)

    # a 3-tuple is returned from get_read_ranges functions
    assert len(read_ranges) == 3

    rr_slices, rr_ranges, rr_scheme_indices = read_ranges

    # - we read four frames, [2, 6)
    # - each frame is divided into 512/32 = 16 tiles in sig dimensions
    # - each tile has depth=2, so we have 4/2*16 = 32 tiles in total
    assert rr_slices.shape == (
        32,  # number of tiles
        2,   # origin and shape
        3,   # indices in flattened dimensions
    )
    # first and last tile slices:
    assert tuple(rr_slices[0, 0]) == (2, 0, 0)
    assert tuple(rr_slices[0, 1]) == (2, 32, 512)
    assert tuple(rr_slices[-1, 0]) == (4, 480, 0)
    assert tuple(rr_slices[-1, 1]) == (2, 32, 512)

    assert rr_ranges.shape == (
        32,   # 32 tiles
        128,  # 128 reads per file per tile (ugh...)
        4     # (file_idx, start, stop, flip)
    )
    # each rr corresponds to a single row:
    row_size = 256 * bits_per_pixel // 8  # row size for a single quadrant, in bytes
    assert set((rr_ranges[:, :, 2] - rr_ranges[:, :, 1]).reshape((-1,))) == {row_size}

    # the offset for the first frame we are reading.
    # we have 3 headers and two full frames ahead of us:
    sig_size_bytes = 512 * 512 * bits_per_pixel // 8
    frame_2_start = 3 * frame_header_bytes + 2 * sig_size_bytes
    q1_offset = 3 * row_size
    q2_offset = 2 * row_size
    q3_offset = 1 * row_size
    q4_offset = 0 * row_size
    stride = 4 * row_size

    # input layout is: [4 | 3 | 2 | 1]
    #
    # output layout is:
    # _________
    # | 1 | 2 |
    # ---------
    # | 3 | 4 |
    # ---------

    # first tile reads only from Q1/Q2:
    # 0, 0 is rr for the first row of Q1
    assert tuple(rr_ranges[0, 0]) == (
        0,  # file index
        frame_2_start + q1_offset + 0 * stride,
        frame_2_start + q1_offset + 0 * stride + row_size,
        0,   # don't flip
    )
    # 0, 1 is rr for the first row of Q2
    assert tuple(rr_ranges[0, 1]) == (
        0,  # file index
        frame_2_start + q2_offset + 0 * stride,
        frame_2_start + q2_offset + 0 * stride + row_size,
        0,   # don't flip
    )
    # 0, 2 is rr for the second row of Q1
    assert tuple(rr_ranges[0, 2]) == (
        0,  # file index
        frame_2_start + q1_offset + 1 * stride,
        frame_2_start + q1_offset + 1 * stride + row_size,
        0,   # don't flip
    )
    # 0, 3 is rr for the second row of Q2
    assert tuple(rr_ranges[0, 3]) == (
        0,  # file index
        frame_2_start + q2_offset + 1 * stride,
        frame_2_start + q2_offset + 1 * stride + row_size,
        0,   # don't flip
    )

    # flip is set in half of the rr's
    assert set(rr_ranges[:8, :, -1].reshape((-1,))) == {0}
    assert set(rr_ranges[8:16, :, -1].reshape((-1,))) == {1}
    # and so on...

    # Let's check out another tile that actually reads from the flipped Q3/Q4
    # (in this case, let's have a look at the last row of the *output*, but still
    # in the first "tile block", so depth starting at frame 2):
    last_tile_idx = 15
    assert tuple(rr_slices[last_tile_idx, 0]) == (2, 480, 0)
    assert tuple(rr_slices[last_tile_idx, 1]) == (2, 32, 512)

    # we have in total a tile depth of 2, which translates to 128
    # read ranges (depth 2, 32 rows per tile, two input rows per output row)
    assert rr_ranges[last_tile_idx].shape == (128, 4)

    # so this means the first input row of Q3 appears at the end of the
    # first half of the read ranges for the full tile.

    # 15, 62 is rr for the first *input* row of Q3, in depth=0 of the tile
    assert tuple(rr_ranges[last_tile_idx, 62]) == (
        0,  # file index
        frame_2_start + q3_offset + 0 * stride,
        frame_2_start + q3_offset + 0 * stride + row_size,
        1,   # flip
    )
    # 15, 63 is rr for the first *input* row of Q4, in depth=0 of the tile
    assert tuple(rr_ranges[last_tile_idx, 63]) == (
        0,  # file index
        frame_2_start + q4_offset + 0 * stride,
        frame_2_start + q4_offset + 0 * stride + row_size,
        1,   # flip
    )

    # for each tile, this is an index into the tiling scheme and indirectly
    # gives us the sig part of the tile slice:
    assert rr_scheme_indices.shape == (32,)
    assert set(rr_scheme_indices) == set(range(16))



import platform

import pytest

from libertem.udf.sum import SumUDF
from libertem.io.dataset.base import IOBackend

from utils import FakeBackend


def test_backend_selection(lt_ctx, default_raw):
    ds = lt_ctx.load(
        "raw",
        path=default_raw._path,
        dtype="float32",
        nav_shape=(16, 16),
        sig_shape=(128, 128),
    )
    p = next(ds.get_partitions())

    expected_backend = 'MMapBackend'
    if platform.system() == 'Windows':
        expected_backend = 'BufferedBackend'

    assert p.get_io_backend().__class__.__name__ == expected_backend


def test_supported_backends():
    backends = IOBackend.get_supported()
    if platform.system() == "Darwin":
        assert backends == ["mmap", "buffered", "fake"]
    else:
        assert backends == ["mmap", "buffered", "direct", "fake"]


def test_load_uses_correct_backend(lt_ctx, default_raw):
    with pytest.raises(RuntimeError):
        ds = lt_ctx.load(
            "raw",
            path=default_raw._path,
            dtype="float32",
            nav_shape=(16, 16),
            sig_shape=(128, 128),
            io_backend=FakeBackend(),
        )
        lt_ctx.run_udf(
            dataset=ds,
            udf=SumUDF(),
        )


def test_auto_uses_correct_backend_hdf5(lt_ctx, hdf5):
    # H5DataSet currently doesn't support alternative I/O backends
    with pytest.raises(ValueError):
        ds = lt_ctx.load(
            "auto",
            path=hdf5.filename,
            ds_path="data",
            io_backend=FakeBackend(),
        )
        lt_ctx.run_udf(
            dataset=ds,
            udf=SumUDF(),
        )



import numpy as np
import pytest
import distributed as dd

from libertem import api
from utils import _naive_mask_apply, _mk_random
from libertem.executor.dask import cluster_spec, DaskJobExecutor
from libertem.executor.pipelined import PipelinedExecutor
from libertem.utils.devices import detect, has_cupy

from utils import DebugDeviceUDF


@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_with_cupy(
    hdf5_ds_1,
    local_cluster_ctx: api.Context,
    pipelined_ctx: api.Context,
    executor
):
    """
    This test only runs if GPUs are detected and cupy is available, and
    makes sure the UDFs run on workers with correct device classes.
    """
    if executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'local_cluster':
        ctx = local_cluster_ctx
    else:
        raise ValueError(f"invalid executor {executor}")
    mask = _mk_random(size=(16, 16))
    d = detect()
    if not d['cudas'] or not d['has_cupy']:
        pytest.skip('this test only runs with a working cupy installation')

    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)

    analysis = ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask]
    )

    # Based on ApplyMasksUDF, which is CuPy-enabled
    hybrid = ctx.run(analysis)
    _ = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cuda', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'cuda', 'numpy')), dataset=hdf5_ds_1)
    cuda_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('cuda', 'numpy')),
        dataset=hdf5_ds_1,
        backends=('cuda',)
    )

    cupy_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('cupy', 'numpy')),
        dataset=hdf5_ds_1,
        backends=('cupy',)
    )
    numpy_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('numpy',)),
        dataset=hdf5_ds_1
    )

    assert np.allclose(
        hybrid.mask_0.raw_data,
        expected
    )
    assert np.all(cuda_only['device_class'].data == 'cuda')
    assert np.all(cupy_only['device_class'].data == 'cuda')
    assert np.all(numpy_only['device_class'].data == 'cpu')


@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_cuda_no_cupy(
    hdf5_ds_1,
    local_cluster_ctx: api.Context,
    pipelined_ctx: api.Context,
    executor
):
    """
    This test only runs if GPUs are detected, but cupy is not available, and
    makes sure the UDFs run on workers with correct device classes.
    """
    if executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'local_cluster':
        ctx = local_cluster_ctx
    else:
        raise ValueError(f"invalid executor {executor}")
    mask = _mk_random(size=(16, 16))
    d = detect()
    if not d['cudas'] or d['has_cupy']:
        pytest.skip('this test only runs with GPU available, but without cupy')

    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)

    analysis = ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask]
    )

    # Based on ApplyMasksUDF, which is CuPy-enabled
    hybrid = ctx.run(analysis)
    _ = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cuda', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'cuda', 'numpy')), dataset=hdf5_ds_1)
    cuda_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('cuda', 'numpy')),
        dataset=hdf5_ds_1,
        backends=('cuda',)
    )
    with pytest.raises(RuntimeError):
        ctx.run_udf(
            udf=DebugDeviceUDF(backends=('cupy', 'numpy')),
            dataset=hdf5_ds_1,
            backends=('cupy',)
        )
    numpy_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('numpy',)),
        dataset=hdf5_ds_1
    )

    assert np.allclose(
        hybrid.mask_0.raw_data,
        expected
    )
    assert np.all(cuda_only['device_class'].data == 'cuda')
    assert np.all(numpy_only['device_class'].data == 'cpu')


@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_no_gpu(
    hdf5_ds_1,
    local_cluster_ctx: api.Context,
    pipelined_ctx: api.Context,
    executor
):
    """
    This test only runs if no GPUs are detected, and makes sure the UDFs run on workers
    with correct device classes
    """
    if executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'local_cluster':
        ctx = local_cluster_ctx
    else:
        raise ValueError(f"invalid executor {executor}")
    mask = _mk_random(size=(16, 16))
    d = detect()
    if d['cudas']:
        pytest.skip('this test only runs without GPUs')

    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)

    analysis = ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask]
    )

    # Based on ApplyMasksUDF, which is CuPy-enabled
    hybrid = ctx.run(analysis)
    _ = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cuda', 'numpy')), dataset=hdf5_ds_1)
    _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy', 'cuda', 'numpy')), dataset=hdf5_ds_1)

    numpy_only = ctx.run_udf(
        udf=DebugDeviceUDF(backends=('numpy',)),
        dataset=hdf5_ds_1
    )
    assert np.allclose(
        hybrid.mask_0.raw_data,
        expected
    )
    assert np.all(numpy_only['device_class'].data == 'cpu')


@pytest.mark.slow
@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_limit_to_cpus(hdf5_ds_1, executor):
    """
    This test explicitly starts executors with only CPU workers, and checks that
    UDFs run on workers with the correct device classes.
    """
    # We don't use all since that might be too many
    # create a Context with only cuda workers:
    cpus = (0, 1)
    if executor == 'pipelined':
        ctx = api.Context.make_with('pipelined', gpus=0, cpus=cpus)
    elif executor == 'local_cluster':
        ctx = api.Context.make_with('dask', gpus=0, cpus=cpus)
    else:
        raise ValueError(f"invalid executor {executor}")
    hdf5_ds_1.set_num_cores(len(cpus))
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)

    with ctx:
        analysis = ctx.create_mask_analysis(
            dataset=hdf5_ds_1, factories=[lambda: mask]
        )
        results = ctx.run(analysis)
        udf_res = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
        # No CuPy resources
        with pytest.raises(RuntimeError):
            _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy',)), dataset=hdf5_ds_1)

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )
    found = {}

    for val in udf_res['device_id'].data[0].values():
        print(val)
        # no CUDA
        assert val["cuda"] is None
        found[val["cpu"]] = True

    for val in udf_res['backend'].data[0].values():
        print(val)
        # no CUDA
        assert 'numpy' in val

    # Each CPU got work. We have to see if this
    # actually works always since this depends on the scheduler behavior
    assert set(found.keys()) == set(cpus)

    assert np.all(udf_res['device_class'].data == 'cpu')
    assert np.allclose(udf_res['on_device'].data, data.sum(axis=(0, 1)))


@pytest.mark.slow
@pytest.mark.skipif(not detect()['cudas'], reason="No CUDA devices")
@pytest.mark.skipif(not has_cupy(), reason="No functional CuPy")
@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_only_cupy(hdf5_ds_1, executor):
    cudas = detect()['cudas']
    # create a Context with only cuda workers:
    if executor == 'pipelined':
        ctx = api.Context.make_with('pipelined', gpus=cudas, cpus=0)
    elif executor == 'local_cluster':
        ctx = api.Context.make_with('dask', gpus=cudas, cpus=0)
    else:
        raise ValueError(f"invalid executor {executor}")
    # Make sure we have enough partitions
    hdf5_ds_1.set_num_cores(len(cudas))
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)

    with ctx:
        # Uses ApplyMasksUDF, which supports CuPy
        analysis = ctx.create_mask_analysis(
            dataset=hdf5_ds_1, factories=[lambda: mask]
        )
        results = ctx.run(analysis)
        udf_res = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
        # No CPU compute resources
        with pytest.raises(RuntimeError):
            _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('numpy',)), dataset=hdf5_ds_1)
        cuda_res = ctx.run_udf(udf=DebugDeviceUDF(backends=('cuda',)), dataset=hdf5_ds_1)

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )

    found = {}

    for val in udf_res['device_id'].data[0].values():
        print(val)
        # no CPU
        assert val["cpu"] is None
        # Register which GPUs got work
        found[val["cuda"]] = True

    for val in cuda_res['device_id'].data[0].values():
        print(val)
        # no CPU
        assert val["cpu"] is None
        # Register which GPUs got work
        found[val["cuda"]] = True

    for val in udf_res['backend'].data[0].values():
        # use CuPy
        print(val)
        assert 'cupy' in val

    for val in cuda_res['backend'].data[0].values():
        # no CuPy, i.e. NumPy
        print(val)
        assert 'numpy' in val

    # Test if each GPU got work. We have to see if this
    # actually works always since this depends on the scheduler behavior
    assert set(found.keys()) == set(cudas)

    assert np.all(udf_res['device_class'].data == 'cuda')
    assert np.allclose(udf_res['on_device'].data, data.sum(axis=(0, 1)))


@pytest.mark.slow
@pytest.mark.skipif(not detect()['cudas'], reason="No CUDA devices")
@pytest.mark.parametrize('executor', ['pipelined', 'local_cluster'])
def test_device_classes_limit_to_cuda(hdf5_ds_1, executor):
    cudas = detect()['cudas']
    # create a Context with only cuda workers (force has_cupy=False):
    if executor == 'pipelined':
        spec = PipelinedExecutor.make_spec(cpus=[], cudas=cudas, has_cupy=False)
        exc = PipelinedExecutor(spec=spec)
    elif executor == 'local_cluster':
        spec = cluster_spec(cpus=(), cudas=cudas, has_cupy=False)
        exc = DaskJobExecutor.make_local(spec=spec)
    else:
        raise ValueError(f"invalid executor {executor}")
    ctx = api.Context(executor=exc)
    # Make sure we have enough partitions
    hdf5_ds_1.set_num_cores(len(cudas))
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]

    with ctx:
        udf_res = ctx.run_udf(udf=DebugDeviceUDF(backends=('cuda', )), dataset=hdf5_ds_1)
        # No CPU compute resources
        with pytest.raises(RuntimeError):
            _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('numpy',)), dataset=hdf5_ds_1)
        # No ndarray (CuPy) resources
        with pytest.raises(RuntimeError):
            _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy',)), dataset=hdf5_ds_1)

    found = {}

    for val in udf_res['device_id'].data[0].values():
        print(val)
        # no CPU
        assert val["cpu"] is None
        # Register which GPUs got work
        found[val["cuda"]] = True

    for val in udf_res['backend'].data[0].values():
        print(val)
        # CUDA, but no CuPy, i.e. use NumPy
        assert 'numpy' in val

    # Test if each GPU got work. We have to see if this
    # actually works always since this depends on the scheduler behavior
    assert set(found.keys()) == set(cudas)

    assert np.all(udf_res['device_class'].data == 'cuda')
    assert np.allclose(udf_res['on_device'].data, data.sum(axis=(0, 1)))


@pytest.mark.slow
def test_use_plain_dask(hdf5_ds_1):
    # We deactivate the resource scheduling and run on a plain dask cluster
    hdf5_ds_1.set_num_cores(2)
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)
    with dd.LocalCluster(n_workers=2, threads_per_worker=1) as cluster:
        client = dd.Client(cluster, set_as_default=False)
        try:
            executor = DaskJobExecutor(client=client)
            ctx = api.Context(executor=executor)
            analysis = ctx.create_mask_analysis(
                dataset=hdf5_ds_1, factories=[lambda: mask]
            )
            results = ctx.run(analysis)
            udf_res = ctx.run_udf(udf=DebugDeviceUDF(), dataset=hdf5_ds_1)
            # Requesting CuPy, which is not available
            with pytest.raises(RuntimeError):
                _ = ctx.run_udf(udf=DebugDeviceUDF(backends=('cupy',)), dataset=hdf5_ds_1)
        finally:
            # to fix "distributed.client - ERROR - Failed to reconnect to scheduler after 10.00 seconds, closing client"  # NOQA
            client.close()

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )

    for val in udf_res['device_id'].data[0].values():
        print(val)
        # no CUDA
        assert val["cuda"] is None
        # Default without worker setup
        assert val["cpu"] == 0

    for val in udf_res['backend'].data[0].values():
        print(val)
        # no CUDA
        assert 'numpy' in val

    assert np.all(udf_res['device_class'].data == 'cpu')
    assert np.allclose(udf_res['on_device'].data, data.sum(axis=(0, 1)))



import numpy as np

from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random

from libertem import masks
from libertem.analysis.sum import SumAnalysis


def test_sum_dataset_tilesize_1(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), num_partitions=32)
    expected = data.sum(axis=(0, 1))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_dataset_tilesize_2(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(8, 16, 16), num_partitions=32)
    expected = data.sum(axis=(0, 1))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_endian(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='>u2')
    dataset = MemoryDataSet(data=data, tileshape=(8, 16, 16), num_partitions=32)
    expected = data.sum(axis=(0, 1))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_signed(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='<i4')
    dataset = MemoryDataSet(data=data, tileshape=(8, 16, 16), num_partitions=32,
                            check_cast=False)
    expected = data.sum(axis=(0, 1))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_timeseries(lt_ctx):
    """
    sum over the first axis of a 3D dataset
    """
    data = _mk_random(size=(16 * 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16, 16),
        num_partitions=32
    )

    # only sum over the first axis:
    expected = data.sum(axis=(0,))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_spectrum_2d_frames(lt_ctx):
    """
    sum over the first two axes of a 3D dataset
    """
    data = _mk_random(size=(16, 16, 16 * 16), dtype='<u2')
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16 * 16),
        num_partitions=32,
        sig_dims=1,
    )

    # sum over the first two axex:
    expected = data.sum(axis=(0, 1))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16 * 16,)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_spectrum_linescan(lt_ctx):
    """
    sum over the first axis of a 2D dataset
    """
    data = _mk_random(size=(16 * 16, 16 * 16), dtype='<u2')
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16 * 16),
        num_partitions=32,
        sig_dims=1,
    )

    # only sum over the first axis:
    expected = data.sum(axis=(0,))

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)

    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16 * 16,)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_hyperspectral(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16, 16),
        num_partitions=32,
        sig_dims=3,
    )

    expected = data.sum(axis=(0, 1))
    analysis = lt_ctx.create_sum_analysis(dataset=dataset)
    results = lt_ctx.run(analysis)

    assert results['intensity'].raw_data.shape == (16, 16, 16)
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_complex(lt_ctx, ds_complex):
    expected = ds_complex.data.sum(axis=(0, 1))
    analysis = lt_ctx.create_sum_analysis(dataset=ds_complex)
    results = lt_ctx.run(analysis)

    assert ds_complex.data.dtype.kind == 'c'
    assert results['intensity_complex'].raw_data.dtype.kind == 'c'

    assert results['intensity'].raw_data.shape == (16, 16)
    assert np.allclose(results['intensity_complex'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, np.abs(expected))
    assert np.allclose(results['intensity'].raw_data, np.abs(expected))


def test_sum_with_roi(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(2, 16, 16), num_partitions=32)

    roi = {
        "shape": "disk",
        "cx": 5,
        "cy": 6,
        "r": 7,
    }
    analysis = SumAnalysis(dataset=dataset, parameters={
        "roi": roi,
    })

    results = lt_ctx.run(analysis)

    mask = masks.circular(roi["cx"], roi["cy"], 16, 16, roi["r"])
    assert mask.shape == (16, 16)
    assert mask[0, 0] == 0
    assert mask[6, 5] == 1
    assert mask.dtype == bool

    # applying the mask flattens the first two dimensions, so we
    # only sum over axis 0 here:
    expected = data[mask, ...].sum(axis=(0,))

    assert expected.shape == (16, 16)
    assert results['intensity'].raw_data.shape == (16, 16)

    # is not equal to results without mask:
    assert not np.allclose(results['intensity'].raw_data, data.sum(axis=(0, 1)))
    # ... but rather like `expected`:
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_zero_roi(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(2, 16, 16), num_partitions=32)

    roi = {
        "shape": "disk",
        "cx": -1,
        "cy": -1,
        "r": 0,
    }
    analysis = SumAnalysis(dataset=dataset, parameters={
        "roi": roi,
    })

    results = lt_ctx.run(analysis)

    mask = masks.circular(roi["cx"], roi["cy"], 16, 16, roi["r"])
    assert mask.shape == (16, 16)
    assert np.count_nonzero(mask) == 0
    assert mask.dtype == bool

    # applying the mask flattens the first two dimensions, so we
    # only sum over axis 0 here:
    expected = data[mask, ...].sum(axis=(0,))

    assert expected.shape == (16, 16)
    assert results['intensity'].raw_data.shape == (16, 16)

    # is not equal to results without mask:
    assert not np.allclose(results['intensity'].raw_data, data.sum(axis=(0, 1)))
    # ... but rather like `expected`:
    assert np.allclose(results['intensity'].raw_data, expected)
    assert np.allclose(results['intensity_lin'].raw_data, expected)


def test_sum_with_crop_frames(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(7, 8, 8),
                            num_partitions=2, sig_dims=2)

    analysis = lt_ctx.create_sum_analysis(dataset=dataset)
    res = lt_ctx.run(analysis)
    print(data.shape, res.intensity.raw_data.shape)
    assert np.allclose(res.intensity.raw_data, np.sum(data, axis=(0, 1)))



from unittest import mock

import numpy as np
import pytest
import matplotlib.pyplot

from libertem.viz.mpl import MPLLive2DPlot
from libertem.udf.sum import SumUDF
from libertem.udf.raw import PickUDF
from libertem.udf.sumsigudf import SumSigUDF


def test_mpl_smoke(monkeypatch, lt_ctx, default_raw):
    udf = SumUDF()
    monkeypatch.setattr(
        matplotlib.pyplot,
        'subplots',
        mock.MagicMock(return_value=(mock.MagicMock(), mock.MagicMock()))
    )
    monkeypatch.setattr(lt_ctx, 'plot_class', MPLLive2DPlot)
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=True)
    matplotlib.pyplot.subplots.assert_called()


def test_mpl_nodisplay(monkeypatch, lt_ctx, default_raw):
    udf = SumUDF()
    monkeypatch.setattr(
        matplotlib.pyplot,
        'subplots',
        mock.MagicMock(return_value=(mock.MagicMock(), mock.MagicMock()))
    )
    plots = [MPLLive2DPlot(dataset=default_raw, udf=udf)]
    with pytest.warns(UserWarning, match="Plot is not displayed, not plotting."):
        lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=plots)


@pytest.mark.parametrize("executor", ["dask", "inline"])
def test_mpl_regression_pick_plot(monkeypatch, local_cluster_ctx, lt_ctx, default_raw, executor):
    if executor == "dask":
        ctx = local_cluster_ctx
    elif executor == "inline":
        ctx = lt_ctx
    udf = PickUDF()
    roi = np.zeros(default_raw.shape.nav, dtype=bool)
    roi[0, 0] = True
    monkeypatch.setattr(
        matplotlib.pyplot,
        'subplots',
        mock.MagicMock(return_value=(mock.MagicMock(), mock.MagicMock()))
    )
    ctx.run_udf(dataset=default_raw, udf=udf, plots=True, roi=roi)


def test_empty(monkeypatch, lt_ctx, default_raw):
    udf = SumSigUDF()
    monkeypatch.setattr(
        matplotlib.pyplot,
        'subplots',
        mock.MagicMock(return_value=(mock.MagicMock(), mock.MagicMock()))
    )
    monkeypatch.setattr(lt_ctx, 'plot_class', MPLLive2DPlot)

    def extract(udf_result, damage):
        return udf_result['intensity'].data, np.zeros_like(damage)

    plot = MPLLive2DPlot(dataset=default_raw, udf=udf, channel=extract)

    with pytest.warns(UserWarning, match='^Plot is not displayed, not plotting'):
        lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=[plot])



from unittest import mock
from libertem.common import Shape, Slice
from libertem.io.dataset.base import TilingScheme


def test_tiling_scheme_methods():
    dataset_shape = Shape((16, 16, 32, 32), sig_dims=2)

    # depth=3 is chosen to not evenly divide the dataset:
    tileshape = Shape((3, 4, 32), sig_dims=2)

    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
    )

    assert len(tiling_scheme) == 8

    expected_slices = [
        Slice(
            origin=(0, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(4, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(8, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(12, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(16, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(20, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(24, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
        Slice(
            origin=(28, 0),
            shape=Shape((4, 32), sig_dims=2),
        ),
    ]
    for i in range(8):
        assert expected_slices[i] == tiling_scheme[i]

    assert [s[1] for s in tiling_scheme.slices] == expected_slices
    assert tiling_scheme.shape == tileshape
    assert tiling_scheme.dataset_shape == dataset_shape
    assert tiling_scheme.depth == 3


def test_adjust_for_partition():
    dataset_shape = Shape((16, 16, 32, 32), sig_dims=2)
    tileshape = Shape((3, 4, 32), sig_dims=2)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
        intent="partition",
    )

    # the mismatch case:
    partition_slice = Slice(
        origin=(0, 0, 0),
        shape=Shape((4, 4, 32), sig_dims=2),
    )
    partition = mock.MagicMock()
    partition.slice = partition_slice

    tiling_scheme_adj = tiling_scheme.adjust_for_partition(partition)
    assert tiling_scheme_adj != tiling_scheme
    assert tiling_scheme_adj.depth == 4

    # the match case:
    partition_slice = Slice(
        origin=(0, 0, 0),
        shape=Shape((3, 4, 32), sig_dims=2),
    )
    partition = mock.MagicMock()
    partition.slice = partition_slice

    tiling_scheme_adj = tiling_scheme.adjust_for_partition(partition)
    assert tiling_scheme_adj == tiling_scheme
    assert tiling_scheme_adj is tiling_scheme
    assert tiling_scheme_adj.depth == 3



import platform

from .backend import IOBackend
from .backend_buffered import BufferedBackendImpl


class DirectBackend(IOBackend, id_="direct"):
    """
    I/O backend using a direct I/O reading strategy. This currently
    works on Linux and Windows, Mac OS X is not yet supported.

    Use this backend if your data is much larger than your RAM, and
    you have fast enough storage (NVMe RAID, for example).
    In these cases, the :code:`MMapBackend` or :code:`BufferedBackend`
    is not efficient, as the system is constantly under memory pressure.  In
    that case, this backend can perform much better.

    Parameters
    ----------
    max_buffer_size : int
        Maximum buffer size, in bytes. This is passed to the tileshape
        negotiation to select the right depth.
    """
    def __init__(self, max_buffer_size=16*1024*1024):
        self._max_buffer_size = max_buffer_size

    @classmethod
    def from_json(cls, msg):
        """
        Construct an instance from the already-decoded `msg`.
        """
        raise NotImplementedError("TODO! implement me!")

    @classmethod
    def platform_supported(cls):
        return platform.system() in ["Linux", "Windows"]

    def get_impl(self):
        return BufferedBackendImpl(
            max_buffer_size=self._max_buffer_size,
            direct_io=True,
        )



import os

import numpy as np
import pytest

from libertem.io.dataset.mrc import MRCDataSet
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.base import BufferedBackend, TilingScheme
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.raw import PickUDF

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

try:
    import mrcfile
except ModuleNotFoundError:
    mrcfile = None


MRC_TESTDATA_PATH = os.path.join(
    get_testdata_path(), 'mrc', '20200821_92978_movie.mrc',
)
HAVE_MRC_TESTDATA = os.path.exists(MRC_TESTDATA_PATH)

pytestmark = pytest.mark.skipif(not HAVE_MRC_TESTDATA, reason="need .mrc testdata")  # NOQA


@pytest.fixture
def default_mrc(lt_ctx):
    ds = lt_ctx.load("mrc", path=MRC_TESTDATA_PATH)
    return ds


@pytest.fixture
def buffered_mrc(lt_ctx):
    # FIXME: not yet supported
    ds = lt_ctx.load("mrc", path=MRC_TESTDATA_PATH, io_backend=BufferedBackend())
    return ds


@pytest.fixture(scope='module')
def default_mrc_raw():
    mrc = mrcfile.open(MRC_TESTDATA_PATH)
    return mrc.data


def test_simple_open(default_mrc):
    assert tuple(default_mrc.shape) == (4, 1024, 1024)


def test_check_valid(default_mrc):
    default_mrc.check_valid()


@pytest.mark.skipif(mrcfile is None, reason="No mrcfile found")
def test_comparison(default_mrc, default_mrc_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_mrc_raw, (-1, *tuple(default_mrc.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_mrc)


@pytest.mark.skipif(mrcfile is None, reason="No mrcfile found")
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_comparison_roi(default_mrc, default_mrc_raw, lt_ctx_fast, as_sparse):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_mrc.shape.nav),
        p=[0.5, 0.5]
    )
    ref_data = default_mrc_raw[roi]
    if as_sparse:
        roi = roi_as_sparse(roi)
    udf = ValidationUDF(reference=ref_data)
    lt_ctx_fast.run_udf(udf=udf, dataset=default_mrc, roi=roi)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction(default_mrc, lt_ctx, with_roi):
    ds = default_mrc

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_detect_1(lt_ctx):
    fpath = MRC_TESTDATA_PATH
    assert MRCDataSet.detect_params(
        path=fpath,
        executor=lt_ctx.executor,
    )["parameters"] == {
        'path': fpath,
        'nav_shape': (2, 2),
        'sig_shape': (1024, 1024)
    }


def test_detect_2(lt_ctx):
    assert MRCDataSet.detect_params(
        path="nofile.someext",
        executor=lt_ctx.executor,
    ) is False


@pytest.mark.dist
def test_mrc_dist(dist_ctx):
    ds = MRCDataSet(path=MRC_TESTDATA_PATH)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    roi = np.random.choice([True, False], size=4)
    results = dist_ctx.run(analysis, roi=roi)
    assert results[0].raw_data.shape == (1024, 1024)


def test_positive_sync_offset(lt_ctx, default_mrc):
    # nav shape 4, we go over the trailing edge
    udf = PickUDF()
    sync_offset = 2

    roi = np.zeros(default_mrc.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[2:4] = True

    ref = lt_ctx.run_udf(dataset=default_mrc, udf=udf, roi=roi)

    ds_with_offset = lt_ctx.load(
        "mrc", path=MRC_TESTDATA_PATH, nav_shape=(2, 2), sync_offset=sync_offset
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    shape = lt_ctx.run_udf(dataset=ds_with_offset, udf=SumSigUDF())

    print(result_with_offset['intensity'].raw_data.shape)
    assert shape['intensity'].data.shape == (2, 2)
    assert np.allclose(
        result_with_offset['intensity'].raw_data[:2],
        ref['intensity'].raw_data
    )


def test_negative_sync_offset(default_mrc, lt_ctx):
    # nav shape 4
    udf = PickUDF()
    sync_offset = -2

    roi = np.zeros(default_mrc.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[:2] = True

    ref = lt_ctx.run_udf(dataset=default_mrc, udf=udf, roi=roi)

    ds_with_offset = lt_ctx.load(
        "mrc", path=MRC_TESTDATA_PATH, nav_shape=(2, 2), sync_offset=sync_offset
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    shape = lt_ctx.run_udf(dataset=ds_with_offset, udf=SumSigUDF())

    print(result_with_offset['intensity'].raw_data.shape)
    assert shape['intensity'].data.shape == (2, 2)
    assert np.allclose(
        result_with_offset['intensity'].raw_data[2:],
        ref['intensity'].raw_data
    )


def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -20

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mrc",
            path=MRC_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-4, 4\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 20

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mrc",
            path=MRC_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-4, 4\), which is \(-image_count, image_count\)"
    )


def test_reshape_nav(lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("mrc", path=MRC_TESTDATA_PATH, nav_shape=(4,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("mrc",  path=MRC_TESTDATA_PATH, nav_shape=(2, 2,))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("mrc",  path=MRC_TESTDATA_PATH, nav_shape=(1, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mrc",
            path=MRC_TESTDATA_PATH,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 1048576"
    )


def test_scheme_too_large(default_mrc):
    partitions = default_mrc.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_mrc.shape.sig),
        sig_dims=default_mrc.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_mrc.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_mrc.shape.sig)


def test_diagnostics(default_mrc):
    assert {"name": "dtype", "value": "float32"} in default_mrc.get_diagnostics()


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("mrc", MRC_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



from typing import (
    Optional, TYPE_CHECKING
)
from typing_extensions import Literal

import numpy as np

from libertem.io.dataset.base import DataSet
from libertem.udf.base import UDF, UDFResultDict

# Base classes for results moved to common for MIT licensing, refs #1031
from libertem.common.analysis import AnalysisResult, AnalysisResultSet

if TYPE_CHECKING:
    from libertem.analysis.helper import GeneratorHelper
    from libertem.web.rpc import ProcedureProtocol
    import numpy.typing as nt


class Analysis:
    """
    Abstract base class for Analysis classes.

    An Analysis is the interface between a UDF and the Web API, and handles
    visualization of partial and full results.

    Passing an instance of an :class:`Analysis` sub-class to
    :meth:`libertem.api.Context.run` will generate an :class:`AnalysisResultSet`.
    The content of this result set is governed by the specific implementation of
    the :code:`Analysis` sub-class.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.7.0
        Removed deprecated methods :code:`get_results` and :code:`get_job`
    """
    TYPE: Literal["UDF"] = "UDF"
    registry: dict[str, "type[Analysis]"] = {}

    def __init__(self, dataset: DataSet, parameters: dict):
        self.dataset = dataset

    def __init_subclass__(cls, id_=None, **kwargs):

        # override id_ with your own id
        # Used to register the subclass
        # https://www.python.org/dev/peps/pep-0487/#subclass-registration

        super().__init_subclass__(**kwargs)
        if id_ is not None:
            cls.registry[id_] = cls

    @classmethod
    def get_analysis_by_type(cls, id_: str) -> type["Analysis"]:
        return cls.registry[id_]

    @classmethod
    def get_template_helper(cls) -> type["GeneratorHelper"]:
        raise NotImplementedError()

    @classmethod
    def get_rpc_definitions(cls) -> dict[str, type["ProcedureProtocol"]]:
        return {}

    async def controller(self, cancel_id, executor, job_is_cancelled, send_results):
        raise NotImplementedError()

    def get_udf_results(
        self, udf_results: UDFResultDict, roi: Optional[np.ndarray],
        damage: "nt.ArrayLike",
    ) -> AnalysisResultSet:
        """
        Convert UDF results to a :code:`AnalysisResultSet`,
        including visualizations.

        Parameters
        ----------
        udf_results
            raw results from the UDF
        roi : numpy.ndarray or None
            Boolean array of the navigation dimension

        Returns
        -------
        list of AnalysisResult
            one or more annotated results
        """
        raise NotImplementedError()

    def get_udf(self) -> UDF:
        """
        Set TYPE='UDF' on the class and implement this method to run a UDF
        from this analysis
        """
        raise NotImplementedError()

    def get_roi(self) -> Optional[np.ndarray]:
        """
        Get the region of interest the UDF should be run on. For example,
        the parameters could describe some geometry, which this method should
        convert to a boolean array. See also: :func:`libertem.analysis.getroi.get_roi`

        Returns
        -------
        numpy.ndarray or None
            region of interest for which we want to run our analysis
        """
        raise NotImplementedError()

    def get_complex_results(self, job_result, key_prefix, title, desc, damage, default_lin=True):
        raise NotImplementedError()

    def get_parameters(self, parameters: dict) -> dict:
        """
        Get analysis parameters. Override to set defaults
        """
        raise NotImplementedError()

    def need_rerun(self, old_params: dict, new_params: dict) -> bool:
        """
        Determine if the analysis needs to be re-run on the data. If not,
        we can just call `get_udf_results` again, for example if the parameters
        only change the visualization.

        Parameters
        ----------
        old_params : Dict
        new_params : Dict

        Returns
        -------
        bool
            True iff the parameter change needs to cause a re-run on the data
        """
        return True


class BaseAnalysis(Analysis):
    def __init__(self, dataset, parameters):
        super().__init__(dataset, parameters)
        self.parameters = self.get_parameters(parameters)
        self.parameters.update(parameters)

        if self.TYPE == 'JOB':
            raise RuntimeError("Job support was removed in 0.7")

    def get_roi(self):
        return None

    def get_complex_results(
            self, job_result, key_prefix, title, desc, damage, default_lin=True):
        from libertem.viz import visualize_simple, CMAP_CIRCULAR_DEFAULT
        damage = damage & np.isfinite(job_result)
        magn = np.abs(job_result)
        angle = np.angle(job_result)
        wheel = CMAP_CIRCULAR_DEFAULT.rgb_from_vector(
            (job_result.real, job_result.imag, 0),
            vmax=np.max(magn[damage])
        )
        return [
            # for compatability, the magnitude has key=key_prefix
            AnalysisResult(
                raw_data=magn,
                visualized=visualize_simple(magn, damage=damage),
                key=key_prefix if default_lin else f'{key_prefix}_lin',
                title="%s [magn]" % title,
                desc="%s [magn]" % desc,
            ),
            AnalysisResult(
                raw_data=magn,
                visualized=visualize_simple(magn, logarithmic=True, damage=damage),
                key=f'{key_prefix}_log' if default_lin else key_prefix,
                title="%s [log(magn)]" % title,
                desc="%s [log(magn)]" % desc,
            ),
            AnalysisResult(
                raw_data=job_result.real,
                visualized=visualize_simple(job_result.real, damage=damage),
                key="%s_real" % key_prefix,
                title="%s [real]" % title,
                desc="%s [real]" % desc,
            ),
            AnalysisResult(
                raw_data=job_result.imag,
                visualized=visualize_simple(job_result.imag, damage=damage),
                key="%s_imag" % key_prefix,
                title="%s [imag]" % title,
                desc="%s [imag]" % desc,
            ),
            AnalysisResult(
                raw_data=angle,
                visualized=visualize_simple(angle, damage=damage),
                key="%s_angle" % key_prefix,
                title="%s [angle]" % title,
                desc="%s [angle]" % desc,
            ),
            AnalysisResult(
                raw_data=job_result,
                visualized=wheel,
                key="%s_complex" % key_prefix,
                title="%s [complex]" % title,
                desc="%s [complex]" % desc,
            ),
        ]

    def get_parameters(self, parameters: dict):
        """
        Get analysis parameters. Override to set defaults
        """
        return parameters


__all__ = ['AnalysisResult', 'AnalysisResultSet', 'Analysis', 'BaseAnalysis']



import numpy as np
import sparse
import inspect

from .masks import SingleMaskAnalysis
from .helper import GeneratorHelper


class PointTemplate(GeneratorHelper):

    short_name = "point"
    api = "create_point_analysis"

    def __init__(self, params):
        self.params = params

    def get_docs(self):
        title = "Point Analysis"
        from libertem.api import Context
        docs_rst = inspect.getdoc(Context.create_point_analysis)
        docs = self.format_docs(title, docs_rst)
        return docs

    def convert_params(self):
        params = ['dataset=ds']
        x = self.params['cx']
        y = self.params['cy']
        params.append(f'x={x}')
        params.append(f'y={y}')
        return ', '.join(params)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(point_result['intensity'])",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class PointMaskAnalysis(SingleMaskAnalysis, id_="APPLY_POINT_SELECTOR"):
    TYPE = 'UDF'

    def get_description(self):
        return "intensity of the integration over the selected point"

    def get_use_sparse(self):
        return True

    def get_mask_factories(self):
        if self.dataset.shape.sig.dims != 2:
            raise ValueError("can only handle 2D signals currently")

        (detector_y, detector_x) = self.dataset.shape.sig

        cx = self.parameters['cx']
        cy = self.parameters['cy']

        sig_shape = self.dataset.shape.sig

        def _point_inner():
            a = sparse.COO(
                data=np.array([1]),
                coords=([int(cy)], [int(cx)]),
                shape=sig_shape
            )
            return a
        return [_point_inner]

    def get_parameters(self, parameters):
        (detector_y, detector_x) = self.dataset.shape.sig

        cx = parameters.get('cx', detector_x / 2)
        cy = parameters.get('cy', detector_y / 2)
        return {
            'cx': cx,
            'cy': cy,
            'mask_count': 1,
            'mask_dtype': np.float32,
        }

    @classmethod
    def get_template_helper(cls):
        return PointTemplate



import os
import sys
import re
import subprocess

import numpy as np
import numba
import pytest
import psutil

from libertem.executor.dask import (
    CommonDaskMixin
)
from libertem.common.scheduler import Worker, WorkerSet
from libertem.executor.dask import cluster_spec, DaskJobExecutor
from libertem.executor.inline import InlineJobExecutor
from libertem.udf.base import UDFRunner
from libertem.udf.base import NoOpUDF
from libertem.udf.sum import SumUDF
from libertem.udf.raw import PickUDF
from libertem.io.dataset.memory import MemoryDataSet
from libertem.api import Context
from libertem.udf import UDF

from utils import _mk_random


def test_task_affinity_1():
    cdm = CommonDaskMixin()

    ws1 = WorkerSet([
        Worker(host='127.0.0.1', name='w1', resources={}, nthreads=1),
        Worker(host='127.0.0.1', name='w2', resources={}, nthreads=1),
        Worker(host='127.0.0.1', name='w3', resources={}, nthreads=1),
        Worker(host='127.0.0.1', name='w4', resources={}, nthreads=1),
    ])
    ws2 = WorkerSet([
        Worker(host='127.0.0.2', name='w5', resources={}, nthreads=1),
        Worker(host='127.0.0.2', name='w6', resources={}, nthreads=1),
        Worker(host='127.0.0.2', name='w7', resources={}, nthreads=1),
        Worker(host='127.0.0.2', name='w8', resources={}, nthreads=1),
    ])
    workers = ws1.extend(ws2)

    assert cdm._task_idx_to_workers(workers, 0) == ws1
    assert cdm._task_idx_to_workers(workers, 1) == ws2
    assert cdm._task_idx_to_workers(workers, 2) == ws1
    assert cdm._task_idx_to_workers(workers, 3) == ws2


@pytest.mark.skipif(os.name == 'nt',
                    reason="Doesn't run on windows")
@pytest.mark.asyncio
async def test_fd_limit(async_executor):
    import resource

    # set soft limit, throws errors but allows to raise it
    # again afterwards:
    proc = psutil.Process()
    oldlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(resource.RLIMIT_NOFILE, (proc.num_fds() + 24, oldlimit[1]))

    print("fds", proc.num_fds())

    try:
        data = _mk_random(size=(1, 16, 16), dtype='<u2')
        dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), num_partitions=1)

        roi = np.ones((1,), dtype=bool)
        udf = PickUDF()

        for i in range(32):
            print(i)
            print(proc.num_fds())

            async for part in UDFRunner([udf]).run_for_dataset_async(
                dataset=dataset,
                executor=async_executor,
                cancel_id="42",
                roi=roi,
            ):
                pass
    finally:
        resource.setrlimit(resource.RLIMIT_NOFILE, oldlimit)


def test_run_each_partition(dask_executor):
    data = _mk_random(size=(16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), num_partitions=16)
    partitions = dataset.get_partitions()

    def fn1(partition):
        return 42

    i = 0
    for result in dask_executor.run_each_partition(partitions, fn1, all_nodes=False):
        i += 1
        assert result == 42
    assert i == 16


def test_run_each_partition_2(dask_executor):
    data = _mk_random(size=(16, 16, 16), dtype='<u2')
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), num_partitions=16)
    partitions = dataset.get_partitions()

    i = 0
    for result in dask_executor.run_each_partition(partitions, lambda p: False, all_nodes=True):
        i += 1
    assert i == 0  # memory dataset doesn't have a defined location, so fn is never run


def test_map_1(dask_executor):
    iterable = [1, 2, 3]
    res = dask_executor.map(lambda x: x**2, iterable)
    assert res == [1, 4, 9]


def test_multiple_clients(local_cluster_url, default_raw):
    ex1 = DaskJobExecutor.connect(local_cluster_url)

    # this creates a second Client, and even though we are setting `set_as_default=False`,
    # this Client is then used by functions like `dd.as_completed`. That is because
    # `set_as_default` only sets the dask scheduler config to "dask.distributed", it does
    # not affect setting the _client_ as the global default `Client`!
    # so any time `as_completed` is called, the `loop` needs to be set correctly, otherwise
    # this may result in strange hangs and crashes
    ex2 = DaskJobExecutor.connect(local_cluster_url)

    udf = SumUDF()

    cx1 = Context(executor=ex1)
    cx1.run_udf(dataset=default_raw, udf=udf)

    ex1.client.close()
    ex2.client.close()


def test_run_each_worker_1(dask_executor):
    def fn1():
        return "some result"
    results = dask_executor.run_each_worker(fn1)
    assert len(results.keys()) >= 1
    assert len(results.keys()) == len(dask_executor.get_available_workers())
    k = next(iter(results))
    result0 = results[k]
    assert result0 == "some result"


@numba.njit(parallel=True)
def something(x):
    result = 0
    for i in numba.prange(23):
        result += x
    return result


class ThreadsPerWorkerUDF(UDF):
    def get_result_buffers(self):
        return {
            'num_threads': self.buffer(kind='nav', dtype=int),
        }

    # Just run once per partition
    def postprocess(self):
        # Actually do Numba multithreading to start threads and
        # freeze maximum thread count
        something(1)

    def process_frame(self, frame):
        assert self.meta.threads_per_worker is not None, \
            "threads_per_worker should be an integer"
        self.results.num_threads[:] = self.meta.threads_per_worker


def test_threads_per_worker(default_raw, dask_executor):
    ctx = Context(executor=dask_executor)
    inline_ctx = Context(executor=InlineJobExecutor())
    res = ctx.run_udf(dataset=default_raw, udf=ThreadsPerWorkerUDF())['num_threads']
    res_inline = inline_ctx.run_udf(dataset=default_raw, udf=ThreadsPerWorkerUDF())['num_threads']
    assert np.allclose(res, 1)
    assert np.allclose(res_inline, psutil.cpu_count(logical=False))


@pytest.mark.slow
def test_threads_per_worker_vanilla(default_raw, monkeypatch):
    old_threads = os.environ.get('NUMBA_NUM_THREADS')
    # Triggers #1053
    monkeypatch.delenv('NUMBA_NUM_THREADS', raising=False)
    ctx = Context()
    assert 'NUMBA_NUM_THREADS' not in os.environ
    # We have to reset it properly since it is set in pytest.ini
    # and Numba will complain if it is changed
    if old_threads:
        os.environ['NUMBA_NUM_THREADS'] = old_threads
    inline_ctx = Context(executor=InlineJobExecutor())
    res = ctx.run_udf(dataset=default_raw, udf=ThreadsPerWorkerUDF())
    res_inline = inline_ctx.run_udf(dataset=default_raw, udf=ThreadsPerWorkerUDF())
    print(res['num_threads'].data)
    assert np.all(res['num_threads'].data == 1)
    print(res_inline['num_threads'].data)
    assert np.all(res_inline['num_threads'].data == psutil.cpu_count(logical=False))


@pytest.mark.slow
@pytest.mark.skipif(
    'DASK_DISTRIBUTED__LOGGING__DISTRIBUTED' in os.environ,
    reason='Setting this value could influence test result',
)
def test_distributed_log_info():
    # Check if there are any distributed.<> - INFO
    # messages in the startup logs
    # Our default config is log_level == WARN
    result = subprocess.run(
        [
            sys.executable,
            '-c',
            'from libertem.api import Context;'
            'ctx = Context.make_with("dask", cpus=1);'
        ],
        timeout=30,
        capture_output=True,
        text=True,
    )

    assert re.search(r'distributed.\S+ - INFO', result.stderr) is None


def test_num_cores(local_cluster_ctx: Context):
    ctx = local_cluster_ctx
    num_cores_ds = ctx.load('memory', data=np.zeros((2, 3, 4, 5)))
    workers = ctx.executor.get_available_workers()
    cpu_count = len(workers.has_cpu())
    gpu_count = len(workers.has_cuda())
    assert num_cores_ds._cores == max(cpu_count, gpu_count)


def test_cluster_spec_cpu_int():
    int_spec = cluster_spec(cpus=4, cudas=tuple(), has_cupy=True)
    range_spec = cluster_spec(cpus=range(4), cudas=tuple(), has_cupy=True)
    assert range_spec == int_spec


def test_cluster_spec_cudas_int():
    spec_n = 4
    cuda_spec = cluster_spec(cpus=tuple(), cudas=spec_n, has_cupy=True)
    num_cudas = 0
    for spec in cuda_spec.values():
        num_cudas += spec.get('options', {}).get('resources', {}).get('CUDA', 0)
    assert num_cudas == spec_n


@pytest.mark.slow
def test_preload(hdf5_ds_1):
    # We don't use all since that might be too many
    cpus = (0, 1)
    hdf5_ds_1.set_num_cores(len(cpus))

    class CheckEnvUDF(NoOpUDF):
        def process_tile(self, tile):
            assert os.environ['LT_TEST_1'] == 'hello'
            assert os.environ['LT_TEST_2'] == 'world'

    preloads = (
        "import os; os.environ['LT_TEST_1'] = 'hello'",
        "import os; os.environ['LT_TEST_2'] = 'world'",
    )

    spec = cluster_spec(cpus=cpus, cudas=(), has_cupy=False, preload=preloads)
    with DaskJobExecutor.make_local(spec=spec) as executor:
        ctx = Context(executor=executor)
        ctx.run_udf(udf=CheckEnvUDF(), dataset=hdf5_ds_1)



import numpy as np

from libertem.masks import _make_circular_mask
from libertem.udf import UDF


class CrystallinityUDF(UDF):
    """
    Determine crystallinity by integration over a ring in the fourier spectrum of each frame.

    Parameters
    ----------

    rad_in: float
        Inner radius in pixels of a ring mask for the integration in Fourier space

    rad_out: float
        Outer radius in pixels of a ring mask for the integration in Fourier space

    real_center: Tuple[float], optional
        (y,x) - pixels, coordinates of a center of a circle for a masking out zero-order peak
        in real space.

    real_rad: float, optional
        Radius in pixels of circle for a masking out zero-order peak in real space.
        If one of real_center or real_rad is missing: the integration will be done without
        masking zero-order peak out.

    Examples
    --------
    >>> cryst_udf = CrystallinityUDF(rad_in=4, rad_out=6, real_center=(8, 8), real_rad=3)
    >>> result = ctx.run_udf(dataset=dataset, udf=cryst_udf)
    >>> np.array(result["intensity"]).shape
    (16, 16)
    """
    def __init__(self, rad_in, rad_out, real_center, real_rad):
        super().__init__(rad_in=rad_in, rad_out=rad_out,
                         real_center=real_center, real_rad=real_rad)

    def get_result_buffers(self):
        return {
            'intensity': self.buffer(
                kind="nav", dtype="float32"
            ),
        }

    def get_task_data(self):
        sigshape = tuple(self.meta.partition_shape.sig)
        rad_in = self.params.rad_in
        rad_out = self.params.rad_out
        real_center = self.params.real_center
        real_rad = self.params.real_rad
        if not (real_center is None or real_rad is None):
            real_mask = 1-1*_make_circular_mask(
                real_center[1], real_center[0], sigshape[1], sigshape[0], real_rad
            )
        else:
            real_mask = None
        fourier_mask_out = 1*_make_circular_mask(
            sigshape[1]*0.5, sigshape[0]*0.5, sigshape[1], sigshape[0], rad_out
        )
        fourier_mask_in = 1*_make_circular_mask(
            sigshape[1]*0.5, sigshape[0]*0.5, sigshape[1], sigshape[0], rad_in
        )
        fourier_mask = np.fft.fftshift(fourier_mask_out - fourier_mask_in)
        half_fourier_mask = fourier_mask[:, :int(fourier_mask.shape[1]*0.5)+1]
        kwargs = {
            'real_mask': real_mask,
            'half_fourier_mask': half_fourier_mask,
        }
        return kwargs

    def process_frame(self, frame):
        h_f_mask = self.task_data.half_fourier_mask
        if self.task_data.real_mask is not None:
            maskedframe = frame*self.task_data.real_mask
        else:
            maskedframe = frame
        self.results.intensity[:] = np.sum(abs(np.fft.rfft2(maskedframe))*h_f_mask)


def run_analysis_crystall(ctx, dataset, rad_in, rad_out, real_center=None, real_rad=None, roi=None,
                          progress=False):
    """
    Return a value after integration of Fourier spectrum for each frame over ring.

    Parameters
    ----------
    ctx : libertem.api.Context
    dataset : libertem.io.dataset.DataSet
        A dataset with 1- or 2-D scan dimensions and 2-D frame dimensions
    rad_in : int
        Inner radius in pixels of a ring mask for the integration in Fourier space
    rad_out : int
        Outer radius in pixels of a ring mask for the integration in Fourier space
    real_center : Tuple[float], optional
        (y,x) - pixels, coordinates of a center of a circle for a masking out zero-order peak
        in real space.
    real_rad : int, optional
        Radius in pixels of circle for a masking out zero-order peak in real space.
        If one of real_center or real_rad is missing: the integration will be done without
        masking zero-order peak out.

    Returns
    -------
    pass_results: dict
        Returns a "crystallinity" value for each frame.
        To return 2-D array use pass_results['intensity'].data

    """
    udf = CrystallinityUDF(
        rad_in=rad_in, rad_out=rad_out, real_center=real_center, real_rad=real_rad
        )
    pass_results = ctx.run_udf(dataset=dataset, udf=udf, roi=roi, progress=progress)

    return pass_results



import os
from typing import Optional, TYPE_CHECKING

import libertem.api as lt
from libertem.io.dataset.base import DataSetException
from libertem.io.dataset.dm_single import SingleDMDataSet
from libertem.common import Shape
from libertem.udf.record import RecordUDF

if TYPE_CHECKING:
    from libertem.api import Context, DataSet


class ConvertTransposedDatasetUDF(RecordUDF):
    def get_method(self):
        return self.UDF_METHOD.PARTITION

    @property
    def _ds_shape(self) -> Shape:
        nav_shape = self.meta.dataset_shape.sig.to_tuple()
        sig_shape = self.meta.dataset_shape.nav.to_tuple()
        return Shape(nav_shape + sig_shape, sig_dims=len(sig_shape))

    @property
    def _memmap_flat_shape(self) -> tuple[int, ...]:
        return (self._ds_shape.nav.size, self._ds_shape.sig.size)

    def process_partition(self, partition):
        # partition will be of shape (n_sig_pix, *ds.shape.nav)
        n_sig_px = partition.shape[0]
        # flatten the nav dimensions
        partition = partition.reshape((n_sig_px, -1))
        # Do the transpose, this is fast but becomes costly
        # the moment we assign into the memmap
        partition = partition.T
        # the LT flat nav origin is actually the sig origin in the memmap
        flat_sig_origin = self.meta.slice.origin[0]
        self.task_data.memmap[
            :, flat_sig_origin:flat_sig_origin + n_sig_px
        ] = partition


def _convert_transposed_ds(
    ctx: 'Context',
    ds: 'DataSet',
    out_path: os.PathLike,
    **run_kwargs,
):
    ctx.run_udf(
        ds,
        ConvertTransposedDatasetUDF(
            out_path,
        ),
        **run_kwargs,
    )


def convert_dm4_transposed(
    dm4_path: os.PathLike,
    out_path: os.PathLike,
    ctx: Optional['Context'] = None,
    num_cpus: Optional[int] = None,
    dataset_index: Optional[int] = None,
    progress: bool = False,
):
    """
    Convenience function to convert a transposed Gatan Digital Micrograph
    (.dm4) STEM dataset into a numpy (.npy) file with standard ordering for
    processing with LiberTEM.

    Transposed .dm4 files are stored in :code:`(sig, nav)` order, i.e.
    all frame values for a given signal pixel are stored as blocks,
    which means that extracting a single frame requires traversal of the
    whole file. LiberTEM requires :code:`(nav, sig)` order for processing
    using the UDF interface, i.e. each frame is stored sequentially.

    .. versionadded:: 0.13.0

    Parameters
    ----------

    dm4_path : PathLike
        The path to the .dm4 file
    out_path : PathLike
        The path to the output .npy file
    ctx : libertem.api.Context, optional
        The Context to use to perform the conversion, by default None
        in which case a Dask-based context will be created (optionally)
        following the :code:`num_cpus` argument.
    num_cpus : int, optional
        When :code:`ctx` is not supplied, this argument limits
        the number of CPUs to perform the conversion. This can be
        important as conversion is a RAM-intensive operation and limiting
        the number of CPUs can help reduce bottlenecking.
    dataset_index : int, optional
        If the .dm4 file contains multiple datasets, this can be used
        to select the dataset to convert
        (see :class:`~libertem.io.dataset.dm_single.SingleDMDataSet`)
        for more information.
    progress : bool, optional
        Whether to display a progress bar during conversion, by default False

    Raises
    ------
    DataSetException
        If the DM4 dataset is not stored as transposed
    ValueError
        If both :code:`ctx` and :code:`num_cpus` are supplied
    """
    if ctx is not None and num_cpus is not None:
        raise ValueError('Either supply a Context or number of cpus to use in conversion')
    elif ctx is None:
        ctx = lt.Context.make_with('dask', cpus=num_cpus)
    ds_meta = SingleDMDataSet._read_metadata(dm4_path, use_ds=dataset_index)
    if ds_meta['c_order']:
        raise DataSetException('The DM4 data is not transposed')
    ds = ctx.load('dm', dm4_path, force_c_order=True, dataset_index=dataset_index)
    return _convert_transposed_ds(ctx, ds, out_path, progress=progress)
